// Wrapping Array is a template class designed to hold a two dimensional
// array of T. It contains an accessor function that allows the user to
// grab an element, as if the array were mapped onto a repeating plane.

#ifndef __WRAPPINGARRAY_H__
#define __WRAPPINGARRAY_H__

#include <vector>

template <class T>
class WrappingArray
{
   public:
      // Constructor creates a wrapping array of a given size.
      WrappingArray(unsigned int _data_x_l, unsigned int _data_y_l)
      {
         assert(_data_x_l != 0);
         assert(_data_y_l != 0);

         data_x_l = _data_x_l;
         data_y_l = _data_y_l;

         // Insert data_x_l * data_y_l copies of standard T constructor.
         data.insert(data.begin(), DataBlockSize(), T());
      }

      // Return element at x_c, y_c on plane.
      T &operator()(int x_c, int y_c)
      {
         int real_x_c = x_c % data_x_l;
         int real_y_c = y_c % data_y_l;

         return ElementAt((unsigned int)real_x_c, (unsigned int)real_y_c);
      }

      const T &operator()(int x_c, int y_c) const
      {
         int real_x_c = x_c % data_x_l;
         int real_y_c = y_c % data_y_l;

         return ElementAt((unsigned int)real_x_c, (unsigned int)real_y_c);
      }

      unsigned int Width() const
      {
         return data_x_l;
      }

      unsigned int Height() const
      {
         return data_y_l;
      }

      void SetAll(const T &proto)
      {
         data.clear();

         data.insert(data.begin(), DataBlockSize(), T(proto));
      }

   private:
      unsigned int data_x_l, data_y_l;

      unsigned int DataBlockSize() const
      {
         return data_x_l * data_y_l;
      }

      // Storage type.
      std::vector<T> data;

      // Returns element within valid grid.
      T &ElementAt(unsigned int x_c, unsigned int y_c)
      {
         unsigned int element = x_c + y_c * data_x_l;

         assert(element < DataBlockSize());
         assert(element < data.size());

         return data.at(element);
      }

      const T &ElementAt(unsigned int x_c, unsigned int y_c) const
      {
         unsigned int element = x_c + y_c * data_x_l;

         assert(element < DataBlockSize());
         assert(element < data.size());

         return data.at(element);
      }
};

#endif // __WRAPPINGARRAY_H__
