// /************************************************************************\
// *                                                                        *
// *  Matt's OpenGL SDL GLEW enabled app                                    *
// *      by Mathew Carr                                                    *
// *                                                                        *
// \************************************************************************/

#define  GLEW_STATIC

#include <GL\glew.h>   // OpenGL Extension Wrangler
#include <GL\wglew.h>  // OpenGL Extension Wrangler (windows!)
#include <SDL\SDL.h>   // libSDL
#include <GL\gl.h>     // OpenGL
#include <GL\glu.h>    // Extra OpenGL functions

#include <windows.h>   // Never know when it might come in handy!

#include "libnoise\noise.h"

#include <cstdio>     // Now for some regular things
#include <cstdlib>
#include <cmath>

#include "mattsdl.h"   // Matt SDL!

#include "VectorMaths.h"
#include "GLVectorMaths.h"
#include "WrappingArray.h"

#define rnd() (((float)rand())/((float)RAND_MAX))

// -----------------------------------------------------------------------------

//const int DISPLAY_X_SIZE = 640;
//const int DISPLAY_Y_SIZE = 480;
const int DISPLAY_X_SIZE = 800;
const int DISPLAY_Y_SIZE = 600;

const float ASPECT_RATIO = (float)DISPLAY_X_SIZE / (float)DISPLAY_Y_SIZE;

const int COLOUR_DEPTH   = 32;

void GenerateLibNoiseFractalTerrain(WrappingArray<float> &output_array)
{
   srand(2);static double f = 0; f+= 0.1;

   noise::module::Perlin p;

   for (unsigned int x = 0; x < output_array.Width()-1; x++)
   {
      for (unsigned int y = 0; y < output_array.Height()-1; y++)
      {
         output_array(x, y) = 0.4*p.GetValue(f+(double)x * 0.004 + 0.31, (double)y * 0.004+0.5, 0.5f);
      }
   }

}

void GenerateDSFractalTerrain(WrappingArray<float> &output_array,
                                             float start_random_range,
                                             float multiplier)
{
   srand(time(NULL)/2);

   // This function generates an interesting landscape
   // using the diamond/square method.

   // Gavin S. P. Miller

   // Only does squares and powers of two for now.
   assert(output_array.Width() == output_array.Height());

   assert((output_array.Width() ==      2)
       || (output_array.Width() ==      4)
       || (output_array.Width() ==      8)
       || (output_array.Width() ==     16)
       || (output_array.Width() ==     32)
       || (output_array.Width() ==     64)
       || (output_array.Width() ==    128)
       || (output_array.Width() ==    256)
       || (output_array.Width() ==    512)
       || (output_array.Width() ==   1024));

   unsigned int iteration = 0;

   const unsigned int array_size = output_array.Width();

   // The size of the 'squares' in the 'diamond step'.
   unsigned int square_size = output_array.Width();

   float random_range = start_random_range;

   while (true)
   {
      random_range *= multiplier;

      unsigned int square_hsize = square_size/2;

      // Perform the 'diamond step' (tetrisect a square by taking its middle point)
      // for each square.

      // FOR EACH SQUARE
      for (unsigned int x_begin = 0; x_begin < array_size; x_begin += square_size)
      {
         for (unsigned int y_begin = 0; y_begin < array_size; y_begin += square_size)
         {
            // Grab the values from the corners of the square.
            float recalled_value = output_array(x_begin              , y_begin              )
                                 + output_array(x_begin              , y_begin + square_size)
                                 + output_array(x_begin + square_size, y_begin              )
                                 + output_array(x_begin + square_size, y_begin + square_size);

            recalled_value /= 4.0f;

            // Distort this value.
            float new_midpoint_value = recalled_value + ((rnd() * random_range * 2.0f) - random_range);

            // Place it
            output_array(x_begin + square_hsize, y_begin + square_hsize) = new_midpoint_value;
         }
      }

      // Perform the 'square step' (tetrisect pairs of diamonds by taking the two points above and right)
      // for each diamond.

      // FOR EACH DIAMOND
      for (unsigned int x_begin = 0; x_begin < array_size; x_begin += square_size)
      {
         for (unsigned int y_begin = 0; y_begin < array_size; y_begin += square_size)
         {
            int diamond_centre_x, diamond_centre_y;

            for (unsigned int position = 0; position < 2; position++)
            {
               if (position == 0)
               {
                  // Diamond to the left of centre (above the square's lower left point).
                  diamond_centre_x = x_begin;
                  diamond_centre_y = y_begin + square_hsize;
               }
               else
               {
                  // Diamond below centre (right of the square's lower left point).
                  diamond_centre_x = x_begin + square_hsize;
                  diamond_centre_y = y_begin;
               }

               // Grab the values from the corners of the diamond. (Above, below, left, right)
               float recalled_value = output_array(diamond_centre_x               , diamond_centre_y + square_hsize)
                                    + output_array(diamond_centre_x               , diamond_centre_y - square_hsize)
                                    + output_array(diamond_centre_x - square_hsize, diamond_centre_y               )
                                    + output_array(diamond_centre_x + square_hsize, diamond_centre_y               );

               recalled_value /= 4.0f;

               // Distort this value.
               float new_midpoint_value = recalled_value + ((rnd() * random_range * 2.0f) - random_range);

               // Place it
               output_array(diamond_centre_x, diamond_centre_y) = new_midpoint_value;
            }
         }
      }

      // Alter variables.
      square_size = square_size / 2;
      iteration++;

      // Breaking conditions
      if (square_size == 1) break;
   }
}

WrappingArray<float> BlahBlur(const WrappingArray<float> &input_array)
{
   WrappingArray<float> na(input_array.Width(), input_array.Height());

   for (int x = 0; x < input_array.Width(); x++)
   {
      for (int y = 0; y < input_array.Height(); y++)
      {
         na(x, y) = (input_array(x - 1, y + 0)
                   + input_array(x + 1, y + 0)
                   + input_array(x + 0, y - 1)
                   + input_array(x + 0, y + 1)
                   + input_array(x + 0, y + 0)) / 5.0f;
      }
   }

   return na;
}
// -----------------------------------------------------------------------------

int main(int argc, char *argv[])
{
   // Start SDL. (video)
   if (SDL_Init(SDL_INIT_VIDEO) < 0)
   {
      // Output error.
      printf("Unable to init SDL: %s\n", SDL_GetError());
      return 1;
   }

   // Make sure SDL cleans up before exit
   atexit(SDL_Quit);

   // Create an OpenGL window.
   SDL_WM_SetCaption("Mathew Carr - SDL OpenGL", NULL);

   SDL_Surface *screen = SDL_SetVideoMode(DISPLAY_X_SIZE,
                                          DISPLAY_Y_SIZE,
                                          COLOUR_DEPTH,
                                          SDL_OPENGL | SDL_HWSURFACE);

   if (!screen)
   {
      // Output error.
      printf("Unable to create video output window: %s\n", SDL_GetError());
      return 1;
   }

   // Enable GLEW!
   glewInit();

   // Set up OpenGL.
   // OpenGL can draw to the entire screen.
   glViewport(0, 0, DISPLAY_X_SIZE, DISPLAY_Y_SIZE);

   // Set the clearing colour to black
   glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

   // Set the depth buffer clearing depth.
   glClearDepth(1.0f);

   // Only draw polygons that are closer than / equal distance to those already drawn.
   glDepthFunc(GL_LEQUAL);

   // Turn on the depth test.
   glEnable(GL_DEPTH_TEST);

   // Smooth shading model
   glShadeModel(GL_SMOOTH);
   // FLAT shading model
   //glShadeModel(GL_FLAT);

   // Lights on
   //glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   glEnable(GL_NORMALIZE);
   glEnable(GL_COLOR_MATERIAL);
   glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

   // Edit the projection matrix - set it to identity - set it to a perspective matrix.
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(37.0f, ASPECT_RATIO, 20.0f, 512.0f);

   // Edit the modelview matrix - set it to identity.
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   float Angle = 0.0f;
   bool toexit = false;

   const unsigned int NO_X_V = 256, NO_Y_V = NO_X_V;

   const float SLOOPS = 0;

   const int DISP_X = NO_X_V + 1, DISP_Y = DISP_X;
   float SCALE = 1.0f;

   WrappingArray<float> values(NO_X_V, NO_Y_V);

   //values(NO_X_V/2, NO_Y_V/2) = 10.0f;

   float multiplier = 0.5f;

#define GenerateDSFractalTerrain(a,b,c) GenerateLibNoiseFractalTerrain(a)

   values.SetAll(0.0f);
   GenerateDSFractalTerrain(values, 2.0f, multiplier);
   values = BlahBlur(values);

   // Main loop.
   while (!toexit) {

//------------------------------------------------------------------------------
      // Start the stopwatch!
      MSDL_SetStartTime();

      // Work all the events.
      MSDL_UpdateEventsArray();
      // Update the MSDL array of keystates.
      MSDL_GetKeyState();

      if (MSDL_keystates[SDLK_o]) SCALE /= 1.02f;
      if (MSDL_keystates[SDLK_p]) SCALE *= 1.02f;

      if (MSDL_keystates[SDLK_q])
      {
         multiplier /= 1.01f;
         values.SetAll(0.0f);
         GenerateDSFractalTerrain(values, 2.0f, multiplier);
   values = BlahBlur(values);
      }
      if (MSDL_keystates[SDLK_w])
      {
         multiplier *= 1.01f;
         values.SetAll(0.0f);
         GenerateDSFractalTerrain(values, 2.0f, multiplier);
   values = BlahBlur(values);
      }

      if (MSDL_triggered_event_flags[SDL_QUIT] || (MSDL_keystates[SDLK_ESCAPE] == 1)) {
         toexit = true;
      }

      // Increase rotation angle.
      //Angle = 45;
      Angle += 0.8f;

      if( Angle > 360.0f ) { Angle -= 360.0f; }

      //------------------------------------------------------------------------

         // Camera.
         glLoadIdentity();

         glTranslatef( 0.0f, 7.0f, -230.0f );

         glScalef(SCALE, SCALE, SCALE);

         glRotatef(40.0f, 1.0f, 0.0f, 0.0f );
         glRotatef(Angle, 0.0f, -1.0f, 0.0f );

         glTranslatef(-50.0f, 0.0f, -50.0f);

         GLfloat lp[4] = {0.0f, 00.0f, 0.0f, 1.0f};
         glLightfv(GL_LIGHT0, GL_POSITION, lp);

         // Clear screen and depth buffer.
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

         glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

         // Draw 3D shapes.
         glBegin(GL_TRIANGLES);

            //glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

            for (int x_v_v = 4+(0 - (SLOOPS * DISP_X/2)); x_v_v <= -4+(DISP_X + (SLOOPS * DISP_X/2)); x_v_v++)
            {
               for (int y_v_v = 4+(0 - (SLOOPS * DISP_Y/2)); y_v_v <= -4+(DISP_Y + (SLOOPS * DISP_Y/2)); y_v_v++)
               {
                  int x_vertex_l = x_v_v + 0;
                  int x_vertex_r = x_v_v + 1;
                  int y_vertex_f = y_v_v + 0;
                  int y_vertex_n = y_v_v + 1;

                  const float HSCALE = 100.0f / ((float)((float)DISP_Y + 1.0f));
                  const float VSCALE = 100.0f;

                  Vector3D vertex_ln = Vector3D(HSCALE * x_vertex_l, VSCALE * values(x_vertex_l, y_vertex_n), HSCALE * y_vertex_n);
                  Vector3D vertex_rn = Vector3D(HSCALE * x_vertex_r, VSCALE * values(x_vertex_r, y_vertex_n), HSCALE * y_vertex_n);
                  Vector3D vertex_lf = Vector3D(HSCALE * x_vertex_l, VSCALE * values(x_vertex_l, y_vertex_f), HSCALE * y_vertex_f);
                  Vector3D vertex_rf = Vector3D(HSCALE * x_vertex_r, VSCALE * values(x_vertex_r, y_vertex_f), HSCALE * y_vertex_f);

                  Vector3D vertex_normal_tri_1 = TriangleNormal(vertex_lf, vertex_rf, vertex_rn);
                  Vector3D vertex_normal_tri_2 = TriangleNormal(vertex_lf, vertex_rn, vertex_ln);

                  float r, g, b;

                  r = ((values(x_vertex_r, y_vertex_f)) * 0.5f) + 0.5f;

                  b=g=r;

                  glColor3f(r, g, b);

                  glNormal3fv(GLfloatVector3D(vertex_normal_tri_1).Values());
                  glVertex3fv(GLfloatVector3D(vertex_lf).Values());
                  glVertex3fv(GLfloatVector3D(vertex_rf).Values());
                  glVertex3fv(GLfloatVector3D(vertex_rn).Values());

                  glColor3f(r, g, b);

                  glNormal3fv(GLfloatVector3D(vertex_normal_tri_1).Values());
                  glVertex3fv(GLfloatVector3D(vertex_lf).Values());
                  glVertex3fv(GLfloatVector3D(vertex_rn).Values());
                  glVertex3fv(GLfloatVector3D(vertex_ln).Values());
               }
            }

         glEnd();

         /*
         // Clear depth buffer.
         glClear(GL_DEPTH_BUFFER_BIT);

         glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

         // Draw 3D shapes.
         glBegin(GL_TRIANGLES);

            glColor4f(1.0f, 1.0f, 0.0f, 1.0f);

            for (int x_v_v = (0 - DISP_X/2); x_v_v <= (DISP_X + DISP_X/2); x_v_v++)
            {
               for (int y_v_v = (0 - DISP_Y/2); y_v_v <= (DISP_Y + DISP_Y/2); y_v_v++)
               {
                  int x_vertex_l = x_v_v + 0;
                  int x_vertex_r = x_v_v + 1;
                  int y_vertex_f = y_v_v + 0;
                  int y_vertex_n = y_v_v + 1;

                  const float HSCALE = 100.0f / ((float)((float)DISP_Y + 1.0f));
                  const float VSCALE = 100.0f;

                  Vector3D vertex_ln = Vector3D(HSCALE * x_vertex_l, VSCALE * values(x_vertex_l, y_vertex_n), HSCALE * y_vertex_n);
                  Vector3D vertex_rn = Vector3D(HSCALE * x_vertex_r, VSCALE * values(x_vertex_r, y_vertex_n), HSCALE * y_vertex_n);
                  Vector3D vertex_lf = Vector3D(HSCALE * x_vertex_l, VSCALE * values(x_vertex_l, y_vertex_f), HSCALE * y_vertex_f);
                  Vector3D vertex_rf = Vector3D(HSCALE * x_vertex_r, VSCALE * values(x_vertex_r, y_vertex_f), HSCALE * y_vertex_f);

                  float r, g, b;

                  r = ((values(x_vertex_r, y_vertex_f)) * 0.5f) + 0.5f;

                  b=g=r;

                  glVertex3fv(GLfloatVector3D(vertex_lf).Values());
                  glVertex3fv(GLfloatVector3D(vertex_rf).Values());
                  glVertex3fv(GLfloatVector3D(vertex_rn).Values());

                  glVertex3fv(GLfloatVector3D(vertex_lf).Values());
                  glVertex3fv(GLfloatVector3D(vertex_rn).Values());
                  glVertex3fv(GLfloatVector3D(vertex_ln).Values());
               }
            }

         glEnd();
         */

      // Swap backbuffer to the screen.
      SDL_GL_SwapBuffers();

      // Limit program speed to 60FPS.
      MSDL_SpinlockUntilStopwatch(15);
   }

//------------------------------------------------------------------------------

   // End program.
   return 0;
}
