// Array2D is a template class designed to hold a two dimensional
// array of T. It contains an accessor function that allows the user to
// grab an element.

#ifndef __ARRAY2D_H__
#define __ARRAY2D_H__

#include "Matrix2DInterface.h"
#include <ctime>

#include <vector>

template <class T>
class Array2D : public Matrix2DInterface<T>
{
   public:
      // Constructor creates a wrapping array of a given size.
      Array2D(unsigned int _data_x_l, unsigned int _data_y_l)
      {
         assert(_data_x_l != 0);
         assert(_data_y_l != 0);

         data_x_l = _data_x_l;
         data_y_l = _data_y_l;

         // Insert data_x_l * data_y_l copies of standard T constructor.
         data.insert(data.begin(), _DataBlockSize(), T());
      }

      // Return element at x_c, y_c.
      virtual T &Element(int _x_c, int _y_c)
      {
         if (!(  (_x_c >= 0)
               &&(_y_c >= 0)
               &&(_x_c < (int)data_x_l)
               &&(_y_c < (int)data_y_l)))
         { throw 0; }
//         assert(_x_c >= 0);
//         assert(_y_c >= 0);
//         assert(_x_c < (int)data_x_l);
//         assert(_y_c < (int)data_y_l);

         return _ElementAt((unsigned int)_x_c, (unsigned int)_y_c);
      }

      virtual const T &Element(int _x_c, int _y_c) const
      {
         if (!(  (_x_c >= 0)
               &&(_y_c >= 0)
               &&(_x_c < (int)data_x_l)
               &&(_y_c < (int)data_y_l)))
         { throw 0; }

//         assert(_x_c >= 0);
//         assert(_y_c >= 0);
//         assert(_x_c < (int)data_x_l);
//         assert(_y_c < (int)data_y_l);

         return _ElementAt((unsigned int)_x_c, (unsigned int)_y_c);
      }

      virtual unsigned int Width() const
      {
         return data_x_l;
      }

      virtual unsigned int Height() const
      {
         return data_y_l;
      }

      virtual void SetAll(const T &proto)
      {
         _SetAll(proto);
      }

      virtual ~Array2D()
      {
      }

   private:
      unsigned int data_x_l, data_y_l;

      unsigned int _DataBlockSize() const
      {
         return data_x_l * data_y_l;
      }

      // Storage type.
      std::vector<T> data;

      // Returns reference to element within valid grid.
      T &_ElementAt(unsigned int _x_c, unsigned int _y_c)
      {
         unsigned int element = _x_c + (_y_c * data_x_l);

         assert(element < _DataBlockSize());
         assert(element < data.size());

         if (!(  (element < _DataBlockSize())
               &&(element < data.size())))
         { throw 0; }

         return data.at(element);
      }

      // Returns reference to element within valid grid.
      const T &_ElementAt(unsigned int _x_c, unsigned int _y_c) const
      {
         unsigned int element = _x_c + (_y_c * data_x_l);

         assert(element < _DataBlockSize());
         assert(element < data.size());

         if (!(  (element < _DataBlockSize())
               &&(element < data.size())))
         { throw 0; }

         return data.at(element);
      }

      void _SetAll(const T &proto)
      {
         data.clear();

         data.insert(data.begin(), _DataBlockSize(), T(proto));
      }
};


void ExportToTGA(const Array2D<float> &array, const char *output_filename = NULL)
{
   char filename[1024] = {0};

   if (output_filename == NULL)
   {
      time_t t = time(NULL);
      strftime(filename, 1023, "%Y-%m-%d %H-%M-%S heightmap.tga", localtime(&t));
      output_filename = filename;
   }

   FILE *output_file = fopen(output_filename, "wb");

   const unsigned short width = array.Width(), height = array.Height();
   const unsigned char * width_char_ptr = (const unsigned char *)& width,
                       *height_char_ptr = (const unsigned char *)&height;

   unsigned char tga_header[18] =
     {0x00,                                     // No extra data
      0x00,                                     // No palette
      0x02,                                     // RGB image
      0x00, 0x00, 0x00, 0x00,                   // No palette data
      0x20,                                     // 32 bit colour...?
      0x00, 0x00, 0x00, 0x00,                   // Start at 0,0
       width_char_ptr[0],  width_char_ptr[1],   // Width
      height_char_ptr[0], height_char_ptr[1],   // Height
      24,                                       // 24 bits per pixel
      0x00};                                     // No bit flips};

   fwrite(tga_header, 18, 1, output_file);

   for (unsigned int y = 0; y < height; y++)
   {
      for (unsigned int x = 0; x < width; x++)
      {
         float fvalue = 255.0f * array.Element(x, y);

         if (fvalue > 255.0f) fvalue = 255.0f;
         if (fvalue <   0.0f) fvalue =   0.0f;

         const unsigned char fchar = (unsigned char)fvalue;

         fprintf(output_file, "%c%c%c", fchar, fchar, fchar);
      }
   }

   fclose(output_file);
}

#endif // __WRAPPINGARRAY2D_H__

