// VectorMaths.cpp/.h    (Original arithmetic.h written by Sean A. Kelly)
// adapted as VectorMaths by Mathew Carr

#ifndef __VECTORMATHS_H__
#define __VECTORMATHS_H__

#define RADIANS_PER_DEGREE (3.14159265f/180.0f)
#define DEGREES_PER_RADIAN (180.0f/3.14159265f)

#ifndef mmax
#define mmax(_a_,_b_) (((_a_) > (_b_)) ? (_a_) : (_b_))
#endif

#ifndef mmin
#define mmin(_a_,_b_) (((_a_) < (_b_)) ? (_a_) : (_b_))
#endif

enum DIMENSION
{
   X = 0,
   Y = 1,
   Z = 2,
   W = 3
};

//Matrix4D structure
struct Matrix4D
{
    //constructor
    Matrix4D();

    // Methods
    void SetToIdentity();
    void SetTranslationPart(float x, float y, float z);
    void SetToRotationX(float radian_x);
    void SetToRotationY(float radian_y);
    void SetToRotationZ(float radian_z);
    void MatrixMultiply(const Matrix4D *m);
    void MatrixMultiply(const Matrix4D &m) {
       MatrixMultiply(&m);
    };
    void MatrixTranspose();

    //Matrix4D cells
    float cell[16];
};
// These matrices are ordered:
// 0  4  8 12
// 1  5  9 13
// 2  6 10 14
// 3  7 11 15

// 3D Vector structure
struct Vector3D
{
	// Constructors
	Vector3D();
	Vector3D(float _x, float _y, float _z);

	// Overloaded operators (Vector3D addition, subtraction, scalar multiplication, division)
	Vector3D  operator+(const Vector3D &b) const;
	Vector3D  operator-(const Vector3D &b) const;
	Vector3D  operator*(float value) const;
	Vector3D  operator/(float value) const;
	Vector3D &operator+=(const Vector3D &b);
	Vector3D &operator-=(const Vector3D &b);
	Vector3D &operator*=(float value);
	Vector3D &operator/=(float value);
   Vector3D  operator-() const;

    // Methods
   float Length() const;
   Vector3D Normal() const;
   void Normalise();
   void MultiplyByMatrix(const Matrix4D *m);

	// Values
	union
	{
	   struct
	   {
	      float x, y, z;
	   };
	   float values[3];
	};
};

Vector3D CrossProduct(const Vector3D *vec1, const Vector3D *vec2);
float DotProduct(const Vector3D *vec1, const Vector3D *vec2);

Vector3D VectorLinearInterpolate(const Vector3D *v1, const Vector3D *v2, float value);
static inline Vector3D VectorLinearInterpolate(const Vector3D &v1, const Vector3D &v2, float value ) {
   return VectorLinearInterpolate(&v1, &v2, value);
}

Vector3D VectorCatmullRomSplineInterpolate(float t, const Vector3D *p0, const Vector3D *p1, const Vector3D *p2, const Vector3D *p3);
static inline Vector3D VectorCatmullRomSplineInterpolate(float t, const Vector3D &p0, const Vector3D &p1, const Vector3D &p2, const Vector3D &p3)
{
   return VectorCatmullRomSplineInterpolate(t, &p0, &p1, &p2, &p3);
}

Vector3D VectorCatmullRomSplineTangent(float t, const Vector3D *p0, const Vector3D *p1, const Vector3D *p2, const Vector3D *p3);
static inline Vector3D VectorCatmullRomSplineTangent(float t, const Vector3D &p0, const Vector3D &p1, const Vector3D &p2, const Vector3D &p3)
{
   return VectorCatmullRomSplineTangent(t, &p0, &p1, &p2, &p3);
}

Vector3D CrossProduct(const Vector3D *vec1, const Vector3D *vec2);
static inline Vector3D CrossProduct(const Vector3D &vec1, const Vector3D &vec2) {
   return CrossProduct(&vec1, &vec2);
}

Vector3D TriangleNormal(const Vector3D &a, const Vector3D &b, const Vector3D &c);
static inline Vector3D TriangleNormal(const Vector3D *a, const Vector3D *b, const Vector3D *c)
{
   return TriangleNormal(*a, *b, *c);
}

static inline Vector3D operator*(float number, const Vector3D &v) {
   return v * number;
}

// Camera struct
struct Camera3D {
   Vector3D ahead; // What direction is this camera pointing?
   Vector3D up;
   Vector3D lateral; // What direction is left to right across the face of the camera?

   Vector3D position;

   // Moves this camera toward another set of specs linearly using a value between 0 and 1.
   Camera3D TendingToward(const Camera3D *c, float value) const;

   Matrix4D AsMatrix() const;

   void BecomeCollinear();
   void DefaultIdentity();
   void NormaliseVectors();
   void RegainSanityAfterChange();
};

/*
//vertex structure
struct vertex_type {
	float x, y, z;                         //position
	float u, v;                            //texture co-ordinate
	float r, g, b, a; 	                  //colour
	float normal_x, normal_y, normal_z;    //normal

   Vector3D get_position() const {
      return Vector3D(x, y, z);
   }

   Vector3D get_normal() const {
      return Vector3D(normal_x, normal_y, normal_z);
   }

   void set_position(Vector3D _p) {
      x = _p.x;
      y = _p.y;
      z = _p.z;
   }

   void set_normal(Vector3D _n) {
      normal_x = _n.x;
      normal_y = _n.y;
      normal_z = _n.z;
   }

   void set_colour(float _r, float _g, float _b, float _a) {
      r = _r;
      g = _g;
      b = _b;
      a = _a;
   }

   void set_colour(float _r, float _g, float _b) {
      r = _r;
      g = _g;
      b = _b;
   }

   void set_colour(Vector3D _v) {
      set_colour(_v.x, _v.y, _v.z);
   }

   void set_texcoords(float _u, float _v) {
      u = _u;
      v = _v;
   }

   vertex_type(float _x        = 0.0f, float _y = 0.0f, float _z = 0.0f,
               float _u        = 0.0f, float _v = 0.0f,
               float _r        = 1.0f, float _g = 1.0f, float _b = 1.0f, float _a = 1.0f,
               float _normal_x = 0.0f, float _normal_y = 0.0f, float _normal_z = 0.0f)
                  : x(_x), y(_y), z(_z),
                  u(_u), v(_v),
                  r(_r), g(_g), b(_b), a(_a),
                  normal_x(_normal_x), normal_y(_normal_y), normal_z(_normal_z)
                     {
   };

   vertex_type(Vector3D _position,
               float _u        = 0.0f, float _v = 0.0f,
               float _r        = 1.0f, float _g = 1.0f, float _b = 1.0f, float _a = 1.0f,
               float _normal_x = 0.0f, float _normal_y = 0.0f, float _normal_z = 0.0f)
                  : x(_position.x), y(_position.y), z(_position.z),
                  u(_u), v(_v),
                  r(_r), g(_g), b(_b), a(_a),
                  normal_x(_normal_x), normal_y(_normal_y), normal_z(_normal_z)
                     {
   };

   vertex_type(Vector3D _position,
               Vector3D _normal,
               float _u        = 0.0f, float _v = 0.0f,
               float _r        = 1.0f, float _g = 1.0f, float _b = 1.0f, float _a = 1.0f)
                  : x(_position.x), y(_position.y), z(_position.z),
                  u(_u), v(_v),
                  r(_r), g(_g), b(_b), a(_a),
                  normal_x(_normal.x), normal_y(_normal.y), normal_z(_normal.z)
                     {
   };

};
*/

#endif // __VECTORMATHS_H__
