// /************************************************************************\
// *                                                                        *
// *  Matt's OpenGL SDL GLEW enabled app                                    *
// *      by Mathew Carr                                                    *
// *                                                                        *
// \************************************************************************/

#define  GLEW_STATIC

#include <GL\glew.h>   // OpenGL Extension Wrangler
#include <GL\wglew.h>  // OpenGL Extension Wrangler (windows!)
#include <SDL\SDL.h>   // libSDL
#include <GL\gl.h>     // OpenGL
#include <GL\glu.h>    // Extra OpenGL functions

#include <windows.h>   // Never know when it might come in handy!

#include <cstdio>      // Now for some regular things
#include <cstdlib>
#include <cmath>

#include "mattsdl.h"   // Matt SDL!

#include "VectorMaths.h"
#include "GLVectorMaths.h"
#include "GLVertexType.h"
#include "WrappingArray2D.h"
#include "Array2D.h"

#include "Colour.h"

#include "WizardBraceletFont.h"

#include "mt19937ar.h"

//#define rnd() (((float)rand())/((float)RAND_MAX))
#define rnd() MT::genrand_realf2()

// -----------------------------------------------------------------------------

const int DISPLAY_X_SIZE = 800;
const int DISPLAY_Y_SIZE = 600;

const float ASPECT_RATIO = (float)DISPLAY_X_SIZE / (float)DISPLAY_Y_SIZE;

const int COLOUR_DEPTH   = 32;

const float ORTHO_PROJECTION_SCREEN_W = 8.0f;
const float ORTHO_PROJECTION_SCREEN_H = 6.0f;
const float ORTHO_PROJECTION_SCREEN_L = -ORTHO_PROJECTION_SCREEN_W * 0.5f;
const float ORTHO_PROJECTION_SCREEN_R =  ORTHO_PROJECTION_SCREEN_W * 0.5f;
const float ORTHO_PROJECTION_SCREEN_T =  ORTHO_PROJECTION_SCREEN_H * 0.5f;
const float ORTHO_PROJECTION_SCREEN_B = -ORTHO_PROJECTION_SCREEN_H * 0.5f;



const float RANDOM_RANGE = 30.0f;

void GenerateDSFractalTerrain(Matrix2DInterface<float> &output_array,
      /* data_square_size is the number of BOXES. (The number of coordinate pins - 1) */
                                             unsigned int data_square_size,
                                             float start_random_range,
                                             float multiplier)
{
   MT::init_by_array((const unsigned long *)"Phoenix Wrighsdasdasder Oven Extravaganza", 10);

   // This function generates an interesting landscape
   // using the diamond/square method.

   // Gavin S. P. Miller

   unsigned int iteration = 0;

   // The size of the 'squares' in the 'diamond step'.
   unsigned int square_size = data_square_size;

   float random_range = start_random_range;

   while (true)
   {
      random_range *= multiplier;

      unsigned int square_hsize = square_size/2;

      // Perform the 'diamond step' (tetrisect a square by taking its middle point)
      // for each square.

      // FOR EACH SQUARE
      for (unsigned int x_begin = 0; x_begin < data_square_size; x_begin += square_size)
      {
         for (unsigned int y_begin = 0; y_begin < data_square_size; y_begin += square_size)
         {
            try
            {
               // Grab the values from the corners of the square.
               float recalled_value = output_array.Element(x_begin              , y_begin              )
                                    + output_array.Element(x_begin              , y_begin + square_size)
                                    + output_array.Element(x_begin + square_size, y_begin              )
                                    + output_array.Element(x_begin + square_size, y_begin + square_size);

               recalled_value /= 4.0f;

               // Distort this value.
               float distortion = ((rnd() * random_range * 2.0f) - random_range);

               // HACK! Centre pin always elevated.
               if (data_square_size == square_size)
               {
                  distortion = RANDOM_RANGE;
               }

               float new_midpoint_value = recalled_value + distortion;

               // Place it
               output_array.Element(x_begin + square_hsize, y_begin + square_hsize) = new_midpoint_value;
            }
            catch (int c)
            {
               // Array access out of bounds. Ignore this one then.
               continue;
            }
         }
      }

      // Perform the 'square step' (tetrisect pairs of diamonds by taking the two points above and right)
      // for each diamond.

      // FOR EACH DIAMOND
      for (unsigned int x_begin = 0; x_begin < data_square_size; x_begin += square_size)
      {
         for (unsigned int y_begin = 0; y_begin < data_square_size; y_begin += square_size)
         {
            int diamond_centre_x, diamond_centre_y;

            // Go through the two diamonds for each square:
            for (unsigned int position = 0; position < 2; position++)
            {
               if (position == 0)
               {
                  // Diamond to the left of centre (above the square's lower left point).
                  diamond_centre_x = x_begin;
                  diamond_centre_y = y_begin + square_hsize;
               }
               else
               {
                  // Diamond below centre (right of the square's lower left point).
                  diamond_centre_x = x_begin + square_hsize;
                  diamond_centre_y = y_begin;
               }

               try
               {
                  // Grab the values from the corners of the diamond. (Above, below, left, right)
                  float recalled_value = output_array.Element(diamond_centre_x               , diamond_centre_y + square_hsize)
                                       + output_array.Element(diamond_centre_x               , diamond_centre_y - square_hsize)
                                       + output_array.Element(diamond_centre_x - square_hsize, diamond_centre_y               )
                                       + output_array.Element(diamond_centre_x + square_hsize, diamond_centre_y               );

                  recalled_value /= 4.0f;

                  // Distort this value.
                  float distortion = ((rnd() * random_range * 2.0f) - random_range);
                  float new_midpoint_value = recalled_value + distortion;

                  // Place it
                  output_array.Element(diamond_centre_x, diamond_centre_y) = new_midpoint_value;
               }
               catch (int c)
               {
                  // Array access out of bounds. Ignore this one then.
                  continue;
               }
            }
         }
      }

      // Alter variables.
      square_size = square_size / 2;
      iteration++;

      // Breaking conditions
      if (square_size == 1) break;
   }
}

WrappingArray2D<float> BlahBlur(const WrappingArray2D<float> &input_array)
{
   WrappingArray2D<float> na(input_array.Width(), input_array.Height());

   for (unsigned int x = 0; x < input_array.Width(); x++)
   {
      for (unsigned int y = 0; y < input_array.Height(); y++)
      {
         na.Element(x, y) = (input_array.Element(x - 1, y + 0)
                           + input_array.Element(x + 1, y + 0)
                           + input_array.Element(x + 0, y - 1)
                           + input_array.Element(x + 0, y + 1)
                           + input_array.Element(x + 0, y + 0)) / 5.0f;
      }
   }

   return na;
}

template <class T>
Array2D<T> Add(
                   const Array2D<T> &a,
                   const Array2D<T> &b)
{
   const unsigned int w = a.Width(), h = a.Height();

   Array2D<T> o(w, h);

   for (unsigned int x = 0; x < w; x++)
   {
      for (unsigned int y = 0; y < h; y++)
      {
         o.Element(x, y) = a.Element(x, y) + b.Element(x, y);
      }
   }

   return o;
}

template <class T>
Array2D<T> Multiply(
                   const Array2D<T> &a,
                   const Array2D<T> &b)
{
   const unsigned int w = a.Width(), h = a.Height();

   Array2D<float> o(w, h);

   for (unsigned int x = 0; x < w; x++)
   {
      for (unsigned int y = 0; y < h; y++)
      {
         o.Element(x, y) = a.Element(x, y) * b.Element(x, y);
      }
   }

   return o;
}

template <class T>
Array2D<T> ClampUpper(
                   const Array2D<T> &a,
                   T v)
{
   const unsigned int w = a.Width(), h = a.Height();

   Array2D<float> o(w, h);

   for (unsigned int x = 0; x < w; x++)
   {
      for (unsigned int y = 0; y < h; y++)
      {
         o.Element(x, y) = mmin(v, a.Element(x, y));
      }
   }

   return o;
}

template <class T>
Array2D<T> ClampLower(
                   const Array2D<T> &a,
                   T v)
{
   const unsigned int w = a.Width(), h = a.Height();

   Array2D<float> o(w, h);

   for (unsigned int x = 0; x < w; x++)
   {
      for (unsigned int y = 0; y < h; y++)
      {
         o.Element(x, y) = mmax(v, a.Element(x, y));
      }
   }

   return o;
}

template <class T>
Array2D<T> ClampBoth(
                   const Array2D<T> &a,
                   T min_value, T max_value)
{
   const unsigned int w = a.Width(), h = a.Height();

   Array2D<float> o(w, h);

   for (unsigned int x = 0; x < w; x++)
   {
      for (unsigned int y = 0; y < h; y++)
      {
         o.Element(x, y) = mmax(min_value, mmin(max_value, a.Element(x, y)));
      }
   }

   return o;
}

template <class T>
Array2D<T> SubtractBfromA(
                   const Array2D<T> &a,
                   const Array2D<T> &b)
{
   const unsigned int w = a.Width(), h = a.Height();

   Array2D<T> o(w, h);

   for (unsigned int x = 0; x < w; x++)
   {
      for (unsigned int y = 0; y < h; y++)
      {
         o.Element(x, y) = a.Element(x, y) - b.Element(x, y);
      }
   }

   return o;
}

   const unsigned int NO_X_V = 128, NO_Y_V = NO_X_V;
   float flompness = 1.0f;

   const float MIN_E = 0.0f, MAX_E = 1.0f;

struct SIGMOID_CONTEXT
{
   float subtract_from_input_first, then_scale_input;

   float a, b, c;

   SIGMOID_CONTEXT(
      float _subtract_from_input_first,
      float _then_scale_input,
      float _a,
      float _b,
      float _c)
         : subtract_from_input_first(_subtract_from_input_first),
           then_scale_input(_then_scale_input),
           a(_a),
           b(_b),
           c(_c)
   {
   }

};

template <typename T>
void SigmoidTransformElement(T &element, const void *_sigmoid_context)
{
   const SIGMOID_CONTEXT *sigmoid_context = (const SIGMOID_CONTEXT *)_sigmoid_context;

   element -= sigmoid_context->subtract_from_input_first;
   element *= sigmoid_context->then_scale_input;

   element = sigmoid_context->a / (1.0f + exp(sigmoid_context->b * (sigmoid_context->c - element)));
}

   // These are the number of BOXES in the grid.
void regenvalues(Array2D<float> &values,
                                             unsigned int data_square_size,
                                             float start_random_range,
                                             float multiplier)
{


   values.SetAll(0.0f);
   GenerateDSFractalTerrain(values, data_square_size, start_random_range, multiplier);
//   values = BlahBlur(values);

// Splat o multiply!

  Array2D<float> multiply_range(NO_X_V+1, NO_Y_V+1);


   for (unsigned int x = 0; x < NO_X_V+1; x++)
   {
      for (unsigned int y = 0; y < NO_Y_V+1; y++)
      {
#define MAX_DIAGONAL_D   sqrt(abs(0 - (NO_X_V/2)) * abs(0 - (NO_X_V/2))   \
                            + abs(0 - (NO_Y_V/2)) * abs(0 - (NO_Y_V/2)))

#define MAX_ORTHOGONAL_D (NO_X_V/2)

         float distance = sqrt(abs(x - (NO_X_V/2)) * abs(x - (NO_X_V/2))
                             + abs(y - (NO_Y_V/2)) * abs(y - (NO_Y_V/2)));

         float multiplier;

         if (distance > MAX_ORTHOGONAL_D)
         {
            multiplier = 0.0f;
         }
         else
         {
            multiplier = pow(1.0f - (distance / MAX_ORTHOGONAL_D), flompness);
         }

         multiply_range.Element(x, y) = multiplier;
      }
   }

   //values = Multiply(values, multiply_range);

   SIGMOID_CONTEXT sigmoid_context = SIGMOID_CONTEXT(0.0f, 1.0f, MAX_E, 3.0f, RANDOM_RANGE / 2.0f);

   values.ForEachUnary(SigmoidTransformElement<float>, &sigmoid_context);

   //values = ClampBoth(values, MIN_E, MAX_E);
}

// -----------------------------------------------------------------------------

int main(int argc, char *argv[])
{
   // Start SDL. (video)
   if (SDL_Init(SDL_INIT_VIDEO) < 0)
   {
      // Output error.
      printf("Unable to init SDL: %s\n", SDL_GetError());
      return 1;
   }

   // Make sure SDL cleans up before exit
   atexit(SDL_Quit);

   // Create an OpenGL window.
   SDL_WM_SetCaption("Mathew Carr - SDL OpenGL", NULL);

   SDL_Surface *screen = SDL_SetVideoMode(DISPLAY_X_SIZE,
                                          DISPLAY_Y_SIZE,
                                          COLOUR_DEPTH,
                                          SDL_OPENGL | SDL_HWSURFACE);

   if (!screen)
   {
      // Output error.
      printf("Unable to create video output window: %s\n", SDL_GetError());
      return 1;
   }

   // Enable GLEW!
   glewInit();

   // Set up OpenGL.
   // OpenGL can draw to the entire screen.
   glViewport(0, 0, DISPLAY_X_SIZE, DISPLAY_Y_SIZE);

   // Set the clearing colour to black
   glClearColor(0.0f, 0.4f, 0.7f, 1.0f);

   // Set the depth buffer clearing depth.
   glClearDepth(1.0f);

   // Only draw polygons that are closer than / equal distance to those already drawn.
   glDepthFunc(GL_LEQUAL);

   // Turn on the depth test.
   glEnable(GL_DEPTH_TEST);

   // Smooth shading model
   glShadeModel(GL_SMOOTH);
   // FLAT shading model
   //glShadeModel(GL_FLAT);

   // Lights on
   //glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   glEnable(GL_NORMALIZE);
   glEnable(GL_COLOR_MATERIAL);
   glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

   float Angle = 0.0f;
   bool toexit = false;

   const float SLOOPS = 0;

   const int DISP_X = NO_X_V + 1, DISP_Y = DISP_X;

   float SCALE = 1.0f;

   //WrappingArray2D<float> values(NO_X_V, NO_Y_V);
   Array2D<float> values(NO_X_V+1, NO_Y_V+1);

   //values(NO_X_V/2, NO_Y_V/2) = 10.0f;

   float multiplier = 0.5f;

   regenvalues(values, NO_X_V, RANDOM_RANGE, multiplier);

   Camera3D current_camera, future_camera;

   current_camera.position = Vector3D( 0.0f,  0.5f,  +1.0f);
   current_camera.lateral  = Vector3D( 1.0f,  0.0f,   0.0f);
   current_camera.up       = Vector3D( 0.0f,  1.0f,   0.0f);
   //current_camera.ahead    = Vector3D( 0.0f,  0.0f,  -1.0f);
   current_camera.ahead = -current_camera.position;

   future_camera = Camera3D(current_camera);

   Uint8 p_mouse_states, c_mouse_states;
   int p_mouse_x, p_mouse_y, c_mouse_x, c_mouse_y;

   //glEnable(GL_CULL_FACE);
   //glCullFace(GL_FRONT);

   WizardBraceletFont::Initialise();

   // Main loop.
   while (!toexit) {

//------------------------------------------------------------------------------
      // Start the stopwatch!
      MSDL_SetStartTime();

      // Work all the events.
      MSDL_UpdateEventsArray();
      // Update the MSDL array of keystates.
      MSDL_GetKeyState();

      if (MSDL_keystates[SDLK_h])
      {
         multiplier /= 1.01f;
         regenvalues(values, NO_X_V, RANDOM_RANGE, multiplier);
      }
      if (MSDL_keystates[SDLK_j])
      {
         multiplier *= 1.01f;
         regenvalues(values, NO_X_V, RANDOM_RANGE, multiplier);
      }

      if (MSDL_keystates[SDLK_n])
      {
         flompness /= 1.01f;
         regenvalues(values, NO_X_V, RANDOM_RANGE, multiplier);
      }
      if (MSDL_keystates[SDLK_m])
      {
         flompness *= 1.01f;
         regenvalues(values, NO_X_V, RANDOM_RANGE, multiplier);
      }

      if (MSDL_triggered_event_flags[SDL_QUIT] || (MSDL_keystates[SDLK_ESCAPE] == 1)) {
         toexit = true;
      }

      if (MSDL_KeyPressed(SDLK_g))
      {
         ExportToTGA(values, NULL);
      }

      MSDL_GetMouseState(&c_mouse_states, &c_mouse_x, &c_mouse_y);

      int x_mouse_motion = c_mouse_x - p_mouse_x;
      int y_mouse_motion = c_mouse_y - p_mouse_y;

      const float MOUSE_MOTION_CAMERA_YAW_TURN_MULTIPLIER = 0.004f;
      const float MOUSE_MOTION_CAMERA_PITCH_TURN_MULTIPLIER = 0.004f;
      const float MOUSE_MOTION_CAMERA_ROLL_TURN_MULTIPLIER = 0.004f;
      const float MOUSE_MOTION_CAMERA_ADVANCE_GO_MULTIPLIER = 0.006f;

      const float KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME = 0.01f;

      if (MSDL_keystates[SDLK_a])
      {
         future_camera.position -= future_camera.lateral * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
      }
      if (MSDL_keystates[SDLK_d])
      {
         future_camera.position += future_camera.lateral * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
      }
      if (MSDL_keystates[SDLK_w])
      {
         future_camera.position += future_camera.ahead * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
      }
      if (MSDL_keystates[SDLK_s])
      {
         future_camera.position -= future_camera.ahead * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
      }
      if (MSDL_keystates[SDLK_q])
      {
         future_camera.position += future_camera.up * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
      }
      if (MSDL_keystates[SDLK_e])
      {
         future_camera.position -= future_camera.up * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
      }

      if (SDL_BUTTON(1) & c_mouse_states)
      {
         if (SDL_BUTTON(3) & c_mouse_states)
         {
            future_camera.up += current_camera.lateral * (((float)x_mouse_motion) * MOUSE_MOTION_CAMERA_ROLL_TURN_MULTIPLIER);
         }
         else
         {
            future_camera.ahead += current_camera.up * (((float)-y_mouse_motion) * MOUSE_MOTION_CAMERA_PITCH_TURN_MULTIPLIER);
            future_camera.ahead += current_camera.lateral * (((float)x_mouse_motion) * MOUSE_MOTION_CAMERA_YAW_TURN_MULTIPLIER);
         }
      }
      else
      {
         if (SDL_BUTTON(3) & c_mouse_states)
         {
            future_camera.position += current_camera.ahead * (((float)-y_mouse_motion) * MOUSE_MOTION_CAMERA_ADVANCE_GO_MULTIPLIER);
         }
      }

      future_camera.RegainSanityAfterChange();

      current_camera = current_camera.TendingToward(&future_camera, 0.2f);

      p_mouse_states = c_mouse_states;
      p_mouse_x = c_mouse_x;
      p_mouse_y = c_mouse_y;

      // Increase rotation angle.
      Angle += 0.1f;

      if( Angle > 360.0f ) { Angle -= 360.0f; }

      //------------------------------------------------------------------------

         // Camera.

         // Edit the projection matrix - set it to identity - set it to a perspective matrix.
         glMatrixMode(GL_PROJECTION);
         glLoadIdentity();
         gluPerspective(37.0f, ASPECT_RATIO, 0.01f, 300.01f);

         // Edit the modelview matrix - set it to identity.
         glMatrixMode(GL_MODELVIEW);
         glLoadIdentity();

         Matrix4D camera_matrix = current_camera.AsMatrix();

         glLoadMatrixf(camera_matrix.cell);

         GLfloat lp[4] = {0.0f, 00.0f, 0.0f, 1.0f};
         glLightfv(GL_LIGHT0, GL_POSITION, lp);

         glScalef(SCALE, SCALE, SCALE);

         // Clear screen and depth buffer.
         glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
         glEnable(GL_DEPTH_TEST);
         glDepthMask(GL_TRUE);

         glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
         glDisable(GL_TEXTURE_2D);
         glBindTexture(GL_TEXTURE_2D, 0);

         // Draw 3D shapes.
         glBegin(GL_TRIANGLES);

            //glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

            for (int x_box = 0; x_box < NO_X_V; x_box++)
            {
               for (int y_box = 0; y_box < NO_Y_V; y_box++)
               {
                  int x_vertex_l = x_box + 0;
                  int x_vertex_r = x_box + 1;
                  int y_vertex_f = y_box + 0;
                  int y_vertex_n = y_box + 1;

                  const float HSCALE = 1.0f / (((float)DISP_Y) - 1.0f);
                  const float VSCALE = 0.2f;

                  Vector3D vertex_ln = Vector3D(HSCALE * (x_vertex_l - NO_X_V / 2.0f), VSCALE * values.Element(x_vertex_l, y_vertex_n), HSCALE * (y_vertex_n - NO_Y_V / 2.0f));
                  Vector3D vertex_rn = Vector3D(HSCALE * (x_vertex_r - NO_X_V / 2.0f), VSCALE * values.Element(x_vertex_r, y_vertex_n), HSCALE * (y_vertex_n - NO_Y_V / 2.0f));
                  Vector3D vertex_lf = Vector3D(HSCALE * (x_vertex_l - NO_X_V / 2.0f), VSCALE * values.Element(x_vertex_l, y_vertex_f), HSCALE * (y_vertex_f - NO_Y_V / 2.0f));
                  Vector3D vertex_rf = Vector3D(HSCALE * (x_vertex_r - NO_X_V / 2.0f), VSCALE * values.Element(x_vertex_r, y_vertex_f), HSCALE * (y_vertex_f - NO_Y_V / 2.0f));

                  Vector3D vertex_normal_tri_1 = TriangleNormal(vertex_lf, vertex_rf, vertex_rn);
                  Vector3D vertex_normal_tri_2 = TriangleNormal(vertex_lf, vertex_rn, vertex_ln);

                  COLOUR c_lf = COLOUR((values.Element(x_vertex_l, y_vertex_f) - MIN_E) / (MAX_E - MIN_E));
                  COLOUR c_rf = COLOUR((values.Element(x_vertex_r, y_vertex_f) - MIN_E) / (MAX_E - MIN_E));
                  COLOUR c_ln = COLOUR((values.Element(x_vertex_l, y_vertex_n) - MIN_E) / (MAX_E - MIN_E));
                  COLOUR c_rn = COLOUR((values.Element(x_vertex_r, y_vertex_n) - MIN_E) / (MAX_E - MIN_E));

                  glNormal3fv(GLfloatVector3D(vertex_normal_tri_1).Values());
                  glColor4fv(c_lf.values); glVertex3fv(GLfloatVector3D(vertex_lf).Values());
                  glColor4fv(c_rf.values); glVertex3fv(GLfloatVector3D(vertex_rf).Values());
                  glColor4fv(c_rn.values); glVertex3fv(GLfloatVector3D(vertex_rn).Values());

                  glNormal3fv(GLfloatVector3D(vertex_normal_tri_1).Values());
                  glColor4fv(c_lf.values); glVertex3fv(GLfloatVector3D(vertex_lf).Values());
                  glColor4fv(c_rn.values); glVertex3fv(GLfloatVector3D(vertex_rn).Values());
                  glColor4fv(c_ln.values); glVertex3fv(GLfloatVector3D(vertex_ln).Values());
               }
            }

         glEnd();
/*
      WizardBraceletFont::DrawStringCentered("-x --- --- +x",
                                     0.0f, 0.3f,
                                     0.03f,
                                     0.03f,
                                     COLOUR::UC(155, 205, 255));

      WizardBraceletFont::DrawString("+y\n|\n|\n|\n|\n|\n|\n-y",
                                     0.0f, 0.7f,
                                     0.03f,
                                     0.03f,
                                     COLOUR::UC(255, 245, 255));

glBindTexture(GL_TEXTURE_2D, 0);
glBegin(GL_TRIANGLES);
glColor3f(1.0f, 1.0f, 0.0f);
glVertex3f(0.0f,  0.7f, 0.0f);
glVertex3f(0.0f, -0.7f, 0.0f);
glVertex3f(0.0f,  0.0f, 0.7f);
glEnd();
*/

      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluOrtho2D(ORTHO_PROJECTION_SCREEN_L, ORTHO_PROJECTION_SCREEN_R, ORTHO_PROJECTION_SCREEN_B, ORTHO_PROJECTION_SCREEN_T);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();

      glDisable(GL_DEPTH_TEST);

      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

      char infoburst[20][200];

      sprintf(infoburst[0], "P: %6.2f, %6.2f, %6.2f", current_camera.position.values[X],
                                                      current_camera.position.values[Y],
                                                      current_camera.position.values[Z]);
      sprintf(infoburst[1], "A: %6.2f, %6.2f, %6.2f", current_camera.ahead.values[X],
                                                      current_camera.ahead.values[Y],
                                                      current_camera.ahead.values[Z]);
      sprintf(infoburst[2], "L: %6.2f, %6.2f, %6.2f", current_camera.lateral.values[X],
                                                      current_camera.lateral.values[Y],
                                                      current_camera.lateral.values[Z]);
      sprintf(infoburst[3], "U: %6.2f, %6.2f, %6.2f", current_camera.up.values[X],
                                                      current_camera.up.values[Y],
                                                      current_camera.up.values[Z]);

      WizardBraceletFont::DrawString(infoburst[0],
                                     ORTHO_PROJECTION_SCREEN_L,
                                     ORTHO_PROJECTION_SCREEN_T - 0 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                     WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                     WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                     COLOUR::UC(255, 255, 255));
      WizardBraceletFont::DrawString(infoburst[1],
                                     ORTHO_PROJECTION_SCREEN_L,
                                     ORTHO_PROJECTION_SCREEN_T - 1 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                     WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                     WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                     COLOUR::UC(155, 255, 255));
      WizardBraceletFont::DrawString(infoburst[2],
                                     ORTHO_PROJECTION_SCREEN_L,
                                     ORTHO_PROJECTION_SCREEN_T - 2 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                     WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                     WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                     COLOUR::UC(255, 155, 255));
      WizardBraceletFont::DrawString(infoburst[3],
                                     ORTHO_PROJECTION_SCREEN_L,
                                     ORTHO_PROJECTION_SCREEN_T - 3 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                     WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                     WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                     COLOUR::UC(255, 255, 155));

      glBindTexture(GL_TEXTURE_2D, 0);

      glDisable(GL_BLEND);

      // Swap backbuffer to the screen.
      SDL_GL_SwapBuffers();

      // Limit program speed to 60FPS.
      MSDL_SpinlockUntilStopwatch(15);
   }

//------------------------------------------------------------------------------

   WizardBraceletFont::Deinitialise();

   // End program.
   return 0;
}
