#include <math.h>
#include "VectorMaths.h"

// VECTOR METHODS

Vector3D::Vector3D()
{
   x = y = z = 0.0f;
}

Vector3D::Vector3D(float _x, float _y, float _z)
   : x(_x), y(_y), z(_z)
{
}

Vector3D Vector3D::operator+(const Vector3D &b) const
{
   return Vector3D(this->x + b.x,   this->y + b.y,   this->z + b.z);
}

Vector3D Vector3D::operator-(const Vector3D &b) const
{
   return Vector3D(this->x - b.x,   this->y - b.y,   this->z - b.z);
}

Vector3D Vector3D::operator*(const Vector3D &b) const /* Weirdo product */
{
   return Vector3D(this->x * b.x, this->y * b.y, this->z * b.z);
}

Vector3D Vector3D::operator*(const Matrix4D *m) const
{
   Vector3D new_vector(*this);

   new_vector.MultiplyByMatrix(m);

   return new_vector;
}

Vector3D Vector3D::operator*(const Matrix4D &m) const
{
   return *this * &m;
}

Vector3D Vector3D::operator*(float value) const
{
   return Vector3D(this->x * value, this->y * value, this->z * value);
}

Vector3D Vector3D::operator/(float value) const
{
   return Vector3D(this->x / value, this->y / value, this->z / value);
}

Vector3D &Vector3D::operator+=(const Vector3D &b)
{
   *this = *this + b;
   return *this;
}

Vector3D &Vector3D::operator-=(const Vector3D &b)
{
   *this = *this - b;
   return *this;
}

Vector3D &Vector3D::operator*=(float value)
{
   *this = *this * value;
   return *this;
}

Vector3D &Vector3D::operator/=(float value)
{
   *this = *this / value;
   return *this;
}

Vector3D Vector3D::operator-() const
{
   return Vector3D(0.0f, 0.0f, 0.0f) - *this;
}

float Vector3D::Length() const
{
    return (sqrt((this->x * this->x)
               + (this->y * this->y)
               + (this->z * this->z)));
}

void Vector3D::Normalise()
{
    float l = Length();

    //if (l < 0.00000001f) l = 0.00000001f;

    this->x /= l;
    this->y /= l;
    this->z /= l;
}

// This returns a copy which is normalised, instead of normalising the original.
Vector3D Vector3D::Normal() const
{
   Vector3D copy = *this;

   copy.Normalise();

   return copy;
}

Vector3D CrossProduct(const Vector3D *vec1, const Vector3D *vec2)
{
   Vector3D r;

   r.x = (vec1->y * vec2->z) - (vec1->z * vec2->y);
	r.y = (vec1->z * vec2->x) - (vec1->x * vec2->z);
	r.z = (vec1->x * vec2->y) - (vec1->y * vec2->x);

	return r;
}

float DotProduct(const Vector3D *vec1, const Vector3D *vec2)
{
    return (vec1->x * vec2->x)
         + (vec1->y * vec2->y)
         + (vec1->z * vec2->z);
}

Vector3D VectorLinearInterpolate(const Vector3D *v1, const Vector3D *v2, float value)
{
   return Vector3D(v1->x + ((v2->x - v1->x) * value),
                   v1->y + ((v2->y - v1->y) * value),
                   v1->z + ((v2->z - v1->z) * value));
}

/*************************************************************************
Vector3D bezier curve interpolation
description:  interpolates between two Vector3Ds using four Vector3D inputs and a
              catmull rom spline.

              This is used for getting POSITION from a curve.
*************************************************************************/

Vector3D VectorCatmullRomSplineInterpolate(float t, const Vector3D *p0, const Vector3D *p1, const Vector3D *p2, const Vector3D *p3)
{
   float t2 = t * t;
   float t3 = t2 * t;

   Vector3D out;

   out.x = 0.5f * ((   2.0f * p1->x )                                              +
                   ( -p0->x + p2->x )                                         * t  +
                   (   2.0f * p0->x -  5.0f * p1->x +  4.0f * p2->x - p3->x ) * t2 +
                   ( -p0->x +  3.0f * p1->x -  3.0f * p2->x + p3->x         ) * t3 );

   out.y = 0.5f * ((   2.0f * p1->y )                                              +
                   ( -p0->y + p2->y )                                         * t  +
                   (   2.0f * p0->y -  5.0f * p1->y +  4.0f * p2->y - p3->y ) * t2 +
                   ( -p0->y +  3.0f * p1->y -  3.0f * p2->y + p3->y         ) * t3 );

   out.z = 0.5f * ((   2.0f * p1->z )                                              +
                   ( -p0->z + p2->z )                                         * t  +
                   (   2.0f * p0->z -  5.0f * p1->z +  4.0f * p2->z - p3->z ) * t2 +
                   ( -p0->z +  3.0f * p1->z -  3.0f * p2->z + p3->z         ) * t3 );

   return out;
}

// This returns the tangent to the spline at t.
Vector3D VectorCatmullRomSplineTangent(float t, const Vector3D *p0, const Vector3D *p1, const Vector3D *p2, const Vector3D *p3)
{
   Vector3D afore = VectorCatmullRomSplineInterpolate(t - 0.01f, p0, p1, p2, p3);
   Vector3D after = VectorCatmullRomSplineInterpolate(t + 0.01f, p0, p1, p2, p3);

   Vector3D delta = after-afore;

   delta.Normalise();

   return delta;
}

void Vector3D::MultiplyByMatrix(const Matrix4D *m)
{
    float xt, yt, zt;

    xt = (x * m->cell[ 0]) + (y * m->cell[ 4]) + (z * m->cell[ 8]);
    yt = (x * m->cell[ 1]) + (y * m->cell[ 5]) + (z * m->cell[ 9]);
    zt = (x * m->cell[ 2]) + (y * m->cell[ 6]) + (z * m->cell[10]);

    x = xt;
    y = yt;
    z = zt;
}

Vector3D TriangleNormal(const Vector3D &a, const Vector3D &b, const Vector3D &c)
{
   Vector3D ba = b - a;
   Vector3D ca = c - a;

   Vector3D normal = CrossProduct(ba, ca);
   normal.Normalise();

   return normal;
}

// MATRIX METHODS

Matrix4D::Matrix4D()
{
   SetToIdentity();
}

void Matrix4D::MatrixMultiply(const Matrix4D *m)
{
   float temp[16];

   //multiply matrices
	temp[ 0] = (cell[ 0] * m->cell[ 0]) + (cell[ 4] * m->cell[ 1]) + (cell[ 8] * m->cell[ 2]) + (cell[12] * m->cell[ 3]); //column 0
	temp[ 1] = (cell[ 1] * m->cell[ 0]) + (cell[ 5] * m->cell[ 1]) + (cell[ 9] * m->cell[ 2]) + (cell[13] * m->cell[ 3]);
	temp[ 2] = (cell[ 2] * m->cell[ 0]) + (cell[ 6] * m->cell[ 1]) + (cell[10] * m->cell[ 2]) + (cell[14] * m->cell[ 3]);
	temp[ 3] = (cell[ 3] * m->cell[ 0]) + (cell[ 7] * m->cell[ 1]) + (cell[11] * m->cell[ 2]) + (cell[15] * m->cell[ 3]);
   temp[ 4] = (cell[ 0] * m->cell[ 4]) + (cell[ 4] * m->cell[ 5]) + (cell[ 8] * m->cell[ 6]) + (cell[12] * m->cell[ 7]); //column 1
	temp[ 5] = (cell[ 1] * m->cell[ 4]) + (cell[ 5] * m->cell[ 5]) + (cell[ 9] * m->cell[ 6]) + (cell[13] * m->cell[ 7]);
	temp[ 6] = (cell[ 2] * m->cell[ 4]) + (cell[ 6] * m->cell[ 5]) + (cell[10] * m->cell[ 6]) + (cell[14] * m->cell[ 7]);
	temp[ 7] = (cell[ 3] * m->cell[ 4]) + (cell[ 7] * m->cell[ 5]) + (cell[11] * m->cell[ 6]) + (cell[15] * m->cell[ 7]);
	temp[ 8] = (cell[ 0] * m->cell[ 8]) + (cell[ 4] * m->cell[ 9]) + (cell[ 8] * m->cell[10]) + (cell[12] * m->cell[11]); //column 2
	temp[ 9] = (cell[ 1] * m->cell[ 8]) + (cell[ 5] * m->cell[ 9]) + (cell[ 9] * m->cell[10]) + (cell[13] * m->cell[11]);
	temp[10] = (cell[ 2] * m->cell[ 8]) + (cell[ 6] * m->cell[ 9]) + (cell[10] * m->cell[10]) + (cell[14] * m->cell[11]);
	temp[11] = (cell[ 3] * m->cell[ 8]) + (cell[ 7] * m->cell[ 9]) + (cell[11] * m->cell[10]) + (cell[15] * m->cell[11]);
	temp[12] = (cell[ 0] * m->cell[12]) + (cell[ 4] * m->cell[13]) + (cell[ 8] * m->cell[14]) + (cell[12] * m->cell[15]); //column 3
	temp[13] = (cell[ 1] * m->cell[12]) + (cell[ 5] * m->cell[13]) + (cell[ 9] * m->cell[14]) + (cell[13] * m->cell[15]);
	temp[14] = (cell[ 2] * m->cell[12]) + (cell[ 6] * m->cell[13]) + (cell[10] * m->cell[14]) + (cell[14] * m->cell[15]);
	temp[15] = (cell[ 3] * m->cell[12]) + (cell[ 7] * m->cell[13]) + (cell[11] * m->cell[14]) + (cell[15] * m->cell[15]);

   //copy Matrix4D result to local Matrix4D
   cell[ 0] = temp[ 0]; cell[ 4] = temp[ 4]; cell[ 8] = temp[ 8]; cell[12] = temp[12];
   cell[ 1] = temp[ 1]; cell[ 5] = temp[ 5]; cell[ 9] = temp[ 9]; cell[13] = temp[13];
   cell[ 2] = temp[ 2]; cell[ 6] = temp[ 6]; cell[10] = temp[10]; cell[14] = temp[14];
	cell[ 3] = temp[ 3]; cell[ 7] = temp[ 7]; cell[11] = temp[11]; cell[15] = temp[15];
}

void Matrix4D::SetTranslationPart(float x, float y, float z)
{
   cell[12] =              x;
   cell[13] =              y;
   cell[14] =              z;
}

void Matrix4D::SetToIdentity()
{
   cell[ 0] =           1.0f; cell[ 4] =           0.0f; cell[ 8] =           0.0f; cell[12] = 0.0f;
   cell[ 1] =           0.0f; cell[ 5] =           1.0f; cell[ 9] =           0.0f; cell[13] = 0.0f;
   cell[ 2] =           0.0f; cell[ 6] =           0.0f; cell[10] =           1.0f; cell[14] = 0.0f;
   cell[ 3] =           0.0f; cell[ 7] =           0.0f; cell[11] =           0.0f; cell[15] = 1.0f;
}

// Matrix4D rotatations
// * transforms a Matrix4D into a matrix representing a rotation
void Matrix4D::SetToRotationX(float radian_x)
{
   cell[ 0] =           1.0f; cell[ 4] =           0.0f; cell[ 8] =           0.0f; cell[12] = 0.0f;
   cell[ 1] =           0.0f; cell[ 5] =  cos(radian_x); cell[ 9] =  sin(radian_x); cell[13] = 0.0f;
   cell[ 2] =           0.0f; cell[ 6] = -sin(radian_x); cell[10] =  cos(radian_x); cell[14] = 0.0f;
   cell[ 3] =           0.0f; cell[ 7] =           0.0f; cell[11] =           0.0f; cell[15] = 1.0f;
}
void Matrix4D::SetToRotationY(float radian_y)
{
   cell[ 0] =  cos(radian_y); cell[ 4] =           0.0f; cell[ 8] = -sin(radian_y); cell[12] = 0.0f;
   cell[ 1] =           0.0f; cell[ 5] =           1.0f; cell[ 9] =           0.0f; cell[13] = 0.0f;
   cell[ 2] =  sin(radian_y); cell[ 6] =           0.0f; cell[10] =  cos(radian_y); cell[14] = 0.0f;
   cell[ 3] =           0.0f; cell[ 7] =           0.0f; cell[11] =           0.0f; cell[15] = 1.0f;
}
void Matrix4D::SetToRotationZ(float radian_z)
{
   cell[ 0] =  cos(radian_z); cell[ 4] =  sin(radian_z); cell[ 8] =           0.0f; cell[12] = 0.0f;
   cell[ 1] = -sin(radian_z); cell[ 5] =  cos(radian_z); cell[ 9] =           0.0f; cell[13] = 0.0f;
   cell[ 2] =           0.0f; cell[ 6] =           0.0f; cell[10] =           1.0f; cell[14] = 0.0f;
   cell[ 3] =           0.0f; cell[ 7] =           0.0f; cell[11] =           0.0f; cell[15] = 1.0f;
}

void Matrix4D::MatrixTranspose()
{
   Matrix4D copy = *this;

   /* cell[ 0]                  */    cell[ 4] = copy.cell[ 1];    cell[ 8] = copy.cell[ 2];       cell[12] = copy.cell[ 3];
      cell[ 1] = copy.cell[ 4];    /* cell[ 5]                  */ cell[ 9] = copy.cell[ 6];       cell[13] = copy.cell[ 7];
      cell[ 2] = copy.cell[ 8];       cell[ 6] = copy.cell[ 9]; /* cell[10]                  */    cell[14] = copy.cell[11];
      cell[ 3] = copy.cell[12];       cell[ 7] = copy.cell[13];    cell[11] = copy.cell[14];    /* cell[15]                  */
}

// Moves this camera toward another set of specs linearly using a value between 0 and 1.
// Returns the resulting camera
Camera3D Camera3D::TendingToward(const Camera3D *c, float value) const
{
   Camera3D new_camera_3d;

   Camera3D * const new_value = &new_camera_3d;

   Vector3D ahead; // What direction is this camera pointing?
   Vector3D up;
   Vector3D lateral; // What direction is left to right across the face of the camera?

   Vector3D position;

   new_value->ahead    = (this->ahead    * (1.0f - value)) + (c->ahead    * (value));
   new_value->lateral  = (this->lateral  * (1.0f - value)) + (c->lateral  * (value));
   new_value->up       = (this->up       * (1.0f - value)) + (c->up       * (value));
   new_value->position = (this->position * (1.0f - value)) + (c->position * (value));

   // Enforce collinearity and normalise vectors!!
   new_camera_3d.RegainSanityAfterChange();

   return new_camera_3d;
}

Matrix4D Camera3D::AsMatrix() const
{
   Matrix4D new_Matrix4D_r, new_Matrix4D_t;

   new_Matrix4D_r.cell[ 0] = lateral.x; new_Matrix4D_r.cell[ 4] = lateral.y; new_Matrix4D_r.cell[ 8] = lateral.z; new_Matrix4D_r.cell[12] = 0;
   new_Matrix4D_r.cell[ 1] =      up.x; new_Matrix4D_r.cell[ 5] =      up.y; new_Matrix4D_r.cell[ 9] =      up.z; new_Matrix4D_r.cell[13] = 0;
   new_Matrix4D_r.cell[ 2] =  -ahead.x; new_Matrix4D_r.cell[ 6] =  -ahead.y; new_Matrix4D_r.cell[10] =  -ahead.z; new_Matrix4D_r.cell[14] = 0;
   new_Matrix4D_r.cell[ 3] = 0;         new_Matrix4D_r.cell[ 7] = 0;         new_Matrix4D_r.cell[11] = 0;         new_Matrix4D_r.cell[15] = 1;

   new_Matrix4D_t.cell[ 0] = 1;         new_Matrix4D_t.cell[ 4] = 0;         new_Matrix4D_t.cell[ 8] = 0;         new_Matrix4D_t.cell[12] = -position.x;
   new_Matrix4D_t.cell[ 1] = 0;         new_Matrix4D_t.cell[ 5] = 1;         new_Matrix4D_t.cell[ 9] = 0;         new_Matrix4D_t.cell[13] = -position.y;
   new_Matrix4D_t.cell[ 2] = 0;         new_Matrix4D_t.cell[ 6] = 0;         new_Matrix4D_t.cell[10] = 1;         new_Matrix4D_t.cell[14] = -position.z;
   new_Matrix4D_t.cell[ 3] = 0;         new_Matrix4D_t.cell[ 7] = 0;         new_Matrix4D_t.cell[11] = 0;         new_Matrix4D_t.cell[15] = 1;

   new_Matrix4D_r.MatrixMultiply(new_Matrix4D_t);
   return new_Matrix4D_r;

   //new_Matrix4D_t.MatrixMultiply(new_Matrix4D_r);
   //return new_Matrix4D_t;
}

void Camera3D::BecomeCollinear()
{
   lateral = CrossProduct(ahead, up);
   //std::printf("Crossing AHEAD-UP: %05.3f, %05.3f, %05.3f\n"
   //            "                   %05.3f, %05.3f, %05.3f = %05.3f, %05.3f, %05.3f\n", ahead.x, ahead.y, ahead.z,
   //                                                                                up.x, up.y, up.z,
   //                                                                                lateral.x, lateral.y, lateral.z);

   up      = CrossProduct(lateral, ahead);
   //std::printf("Crossing LATERAL-AH: %05.3f, %05.3f, %05.3f\n"
   //            "                     %05.3f, %05.3f, %05.3f = %05.3f, %05.3f, %05.3f\n", lateral.x, lateral.y, lateral.z,
   //                                                                                ahead.x, ahead.y, ahead.z,
   //                                                                                up.x, up.y, up.z);
}

void Camera3D::DefaultIdentity()
{
       ahead.x = 0.0f;
       ahead.y = 0.0f;
       ahead.z = 1.0f;

          up.x = 0.0f;
          up.y = 1.0f;
          up.z = 0.0f;

     lateral.x = 1.0f;
     lateral.y = 0.0f;
     lateral.z = 0.0f;

    position.x = 0.0f;
    position.y = 0.0f;
    position.z = 0.0f;
}

void Camera3D::NormaliseVectors()
{
     ahead.Normalise();
        up.Normalise();
   lateral.Normalise();
}

void Camera3D::RegainSanityAfterChange()
{
   NormaliseVectors();
   BecomeCollinear();
}
