#include "WizardBraceletFont.h"

#include <vector>

#include "skMrDTGA.h"

namespace WizardBraceletFont
{
   const char *FONT_TEXTURE_FILENAME = "font.tga";

   bool font_loaded = false;

   GLuint font_texture_number;

   // Try to load the WizardBracelet Font

   // Return true on success / already loaded.
   // Return false on failure.
   bool Initialise()
   {
      // Try to load the wizard bracelet font.

      // If the font already exists, that's good
      // automatic success.
      if (font_loaded)
      {
         return true;
      }

      // Try to load the font.
      bool success = SK_MrD_LoadTGA(FONT_TEXTURE_FILENAME, font_texture_number);

      // False on failure.
      if (!success)
      {
         return false;
      }

      // Success.
      font_loaded = true;

      return true;
   }

   void Deinitialise()
   {
      if (font_loaded)
      {
         glDeleteTextures(1, &font_texture_number);

         font_texture_number = 0;
      }

      font_loaded = false;
   }

   // What is the layout of the character grid
   const int   CHARACTERS_PER_ROW    = 16;
   const int   CHARACTERS_PER_COLUMN = 16;

   const float TEXTURE_INCREMENT_X_PER_CHARACTER = ( 1.0f / ((float)CHARACTERS_PER_ROW   ));
   const float TEXTURE_INCREMENT_Y_PER_CHARACTER = (-1.0f / ((float)CHARACTERS_PER_COLUMN));

   const float DEFAULT_CHARACTER_SIZE    = 0.20f;
   const float DEFAULT_CHARACTER_SPACING = 0.58f;
   const float DEFAULT_LINE_SPACING      = 0.93f;

   // Renders the given null terminated string to the position.
   // \n and \r are treated as newlines.

   // It fiddles with the render states and texture states.
   // Make sure to set these explicitly later before
   // attempting further drawing!
   void DrawString(const char *characters,
                   float l_x,                      // X position of left side of string.
                   float t_y,                      // Y position of top of string.
                   float x_size,                   // X size of rendered character.
                   float y_size,                   // Y size of rendered character.
                   COLOUR colour,                  // Colour to apply to text.
                   float x_character_spacing,      // Spacing between characters. (in multiples of character size)
                   float y_line_spacing)           // Spacing between lines. (in multiples of character size)
   {
      // Do nothing if the font isn't loaded.
      if (!font_loaded)
      {
         return;
      }

      // Can't draw empty string.
      if (strlen(characters) == 0)
      {
         return;
      }

      // Initalise positioning counters.
      float current_l_x = l_x;
      float current_r_x = l_x + x_size;

      float current_t_y = t_y;
      float current_b_y = t_y - y_size;

      // This function works by populating a vector of CUSTOMVERTEX
      // with the vertices necessary to render the string given.

      int x_in_grid = 0, y_in_grid = 0;

      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D, font_texture_number);

      glBegin(GL_QUADS);

      // Process all characters until 0 character.
      for (; *characters != 0; characters++)
      {
         if ((*characters == '\n')
          || (*characters == '\r'))
         {
            // Reset the X values, advance the Y values.
            current_l_x = l_x;
            current_r_x = l_x + x_size;

            current_t_y -= y_line_spacing * y_size;
            current_b_y -= y_line_spacing * y_size;

            continue;
         }

         // Do visible character calculations
         // for all characters EXCEPT space.
         if ((*characters != ' ')
          && (*characters  >  32))
         {
            // Printable characters get their position in the
            // texture grid located:
            // This variable treats the grid as linear.
            // Default characters are question marks.
            unsigned int character_position = *characters;

            // Find the position in the grid.
            x_in_grid = character_position % CHARACTERS_PER_ROW;
            y_in_grid = character_position / CHARACTERS_PER_ROW;

            // Calculate texture coordinates for that character.
            float texture_l_coordinate = ((float)( x_in_grid + 0)) * TEXTURE_INCREMENT_X_PER_CHARACTER;
            float texture_r_coordinate = ((float)( x_in_grid + 1)) * TEXTURE_INCREMENT_X_PER_CHARACTER;
            float texture_t_coordinate = ((float)( y_in_grid + 0)) * TEXTURE_INCREMENT_Y_PER_CHARACTER;
            float texture_b_coordinate = ((float)( y_in_grid + 1)) * TEXTURE_INCREMENT_Y_PER_CHARACTER;

            const float IGNORE_Z = 0.01f;

            // Create vertices.
            glColor4fv(colour.values);

            glTexCoord2d(texture_l_coordinate, texture_t_coordinate);
            glVertex3f(current_l_x, current_t_y, IGNORE_Z);
            glTexCoord2d(texture_r_coordinate, texture_t_coordinate);
            glVertex3f(current_r_x, current_t_y, IGNORE_Z);
            glTexCoord2d(texture_r_coordinate, texture_b_coordinate);
            glVertex3f(current_r_x, current_b_y, IGNORE_Z);
            glTexCoord2d(texture_l_coordinate, texture_b_coordinate);
            glVertex3f(current_l_x, current_b_y, IGNORE_Z);
         }

         // Advance the vertex positioning values.
         current_l_x += (x_character_spacing * x_size);
         current_r_x += (x_character_spacing * x_size);
      }

      glEnd();
   }
};
