// /************************************************************************\
// *                                                                        *
// *  Matt's OpenGL SDL GLEW enabled app                                    *
// *      by Mathew Carr                                                    *
// *                                                                        *
// \************************************************************************/

#define  GLEW_STATIC

#include <GL\glew.h>   // OpenGL Extension Wrangler
#include <GL\wglew.h>  // OpenGL Extension Wrangler (windows!)
#include <SDL\SDL.h>   // libSDL
#include <GL\gl.h>     // OpenGL
#include <GL\glu.h>    // Extra OpenGL functions

#include <windows.h>   // Never know when it might come in handy!

#include <cstdio>      // Now for some regular things
#include <cstdlib>
#include <cmath>

#include "mattsdl.h"   // Matt SDL!

#include "VectorMaths.h"
#include "GLVectorMaths.h"
#include "GLVertexType.h"
#include "WrappingArray2D.h"
#include "Array2D.h"

#include "Colour.h"

#include "WizardBraceletFont.h"

#include "mt19937ar.h"

//#define rnd() (((float)rand())/((float)RAND_MAX))
#define rnd() MT::genrand_realf2()

// -----------------------------------------------------------------------------

SDL_Surface *screensurface;

const int DISPLAY_X_SIZE = 800;
const int DISPLAY_Y_SIZE = 600;

const float ASPECT_RATIO = (float)DISPLAY_X_SIZE / (float)DISPLAY_Y_SIZE;

const int COLOUR_DEPTH   = 32;

const float ORTHO_PROJECTION_SCREEN_W = 8.0f;
const float ORTHO_PROJECTION_SCREEN_H = 6.0f;
const float ORTHO_PROJECTION_SCREEN_L = -ORTHO_PROJECTION_SCREEN_W * 0.5f;
const float ORTHO_PROJECTION_SCREEN_R =  ORTHO_PROJECTION_SCREEN_W * 0.5f;
const float ORTHO_PROJECTION_SCREEN_T =  ORTHO_PROJECTION_SCREEN_H * 0.5f;
const float ORTHO_PROJECTION_SCREEN_B = -ORTHO_PROJECTION_SCREEN_H * 0.5f;

unsigned int power_of_two(unsigned int power)
{
   unsigned int x = 1;
   for (; power > 0; power--) x*=2;
   return x;
}

const unsigned int NO_SEED_STRING_CHARACTERS = 100;

struct ENVIRONMENT_GENERATION_PARAMETERS
{
   char seed_string[NO_SEED_STRING_CHARACTERS];

   unsigned int n_value;

   unsigned int no_x_cells;
   unsigned int no_y_cells;

   float water_level;

   float starting_deviation_range_value;
   float deviation_multiplier;

   bool radial_falloff_enable;
   float radial_falloff_value;

   bool sigmoid_curve_enable;
   float sigmoid_curve_a_parameter_value;
   float sigmoid_curve_b_parameter_value;
   float sigmoid_curve_c_parameter_value;

   float minimum_colour_height_value;
   float maximum_colour_height_value;

   unsigned int terrain_colour_number;
};

struct FIXED_VALUE_ARRAY_ENTRY
{
   float value;
   bool fixed;

   FIXED_VALUE_ARRAY_ENTRY()
    : value(0.0f), fixed(false)
    {}
};

struct PRESENTATION_PARAMETERS
{
   bool lighting_enabled;
   bool semitransparent_water_active;
   bool smooth_shaded_terrain_active;
};

float SCALE = 1.0f;
const float VSCALE = 0.4f;

const float RANDOM_RANGE = 2.0f;

void GenerateDSFractalTerrain(Matrix2DInterface<float> &output_array,
      /* data_square_size is the number of BOXES. (The number of coordinate pins - 1) */
                              const ENVIRONMENT_GENERATION_PARAMETERS &environment_parameters,
                              const Matrix2DInterface<FIXED_VALUE_ARRAY_ENTRY> *fixed_values = NULL)
{
   MT::init_by_array((const unsigned long *)environment_parameters.seed_string, NO_SEED_STRING_CHARACTERS/4);

   // This function generates an interesting landscape
   // using the diamond/square method.

   // Benoit Mandelbrot
   // Gavin S. P. Miller

   unsigned int iteration = 0;

   // The size of the 'squares' in the 'diamond step'.
   const unsigned int &data_square_size = environment_parameters.no_x_cells;
   unsigned int square_size = environment_parameters.no_x_cells; /* assumes square */

   float random_range = environment_parameters.starting_deviation_range_value;

   while (true)
   {
      unsigned int square_hsize = square_size/2;

      // Perform the 'diamond step' (tetrisect a square by taking its middle point)
      // for each square.

      // FOR EACH SQUARE
      for (unsigned int x_begin = 0; x_begin < data_square_size; x_begin += square_size)
      {
         for (unsigned int y_begin = 0; y_begin < data_square_size; y_begin += square_size)
         {
            int x_setting_element = x_begin + square_hsize;
            int y_setting_element = y_begin + square_hsize;

            bool overridden_by_fixed_array = false;

            if (fixed_values != NULL)
            {
               if (fixed_values->Element(x_setting_element, y_setting_element).fixed)
               {
                  output_array.Element(x_setting_element, y_setting_element) =
                    fixed_values->Element(x_setting_element, y_setting_element).value;

                  overridden_by_fixed_array = true;
               }
            }

            if (overridden_by_fixed_array == false)
            {
               try
               {
                  // Grab the values from the corners of the square.
                  float recalled_value = output_array.Element(x_begin              , y_begin              )
                                       + output_array.Element(x_begin              , y_begin + square_size)
                                       + output_array.Element(x_begin + square_size, y_begin              )
                                       + output_array.Element(x_begin + square_size, y_begin + square_size);

                  recalled_value /= 4.0f;

                  // Distort this value.
                  float distortion = ((rnd() * random_range * 2.0f) - random_range);

                  float new_midpoint_value = recalled_value + distortion;

                  // Place it
                  output_array.Element(x_setting_element, y_setting_element) = new_midpoint_value;
               }
               catch (int c)
               {
                  // Array access out of bounds. Ignore this one then.
                  continue;
               }
            }
         }
      }

      // Perform the 'square step' (tetrisect pairs of diamonds by taking the two points above and right)
      // for each diamond.

      // FOR EACH DIAMOND
      for (unsigned int x_begin = 0; x_begin < data_square_size; x_begin += square_size)
      {
         for (unsigned int y_begin = 0; y_begin < data_square_size; y_begin += square_size)
         {
            int diamond_centre_x, diamond_centre_y;

            // Go through the two diamonds for each square:
            for (unsigned int position = 0; position < 2; position++)
            {
               if (position == 0)
               {
                  // Diamond to the left of centre (above the square's lower left point).
                  diamond_centre_x = x_begin;
                  diamond_centre_y = y_begin + square_hsize;
               }
               else
               {
                  // Diamond below centre (right of the square's lower left point).
                  diamond_centre_x = x_begin + square_hsize;
                  diamond_centre_y = y_begin;
               }


               bool overridden_by_fixed_array = false;

               if (fixed_values != NULL)
               {
                  if (fixed_values->Element(diamond_centre_x, diamond_centre_y).fixed)
                  {
                     output_array.Element(diamond_centre_x, diamond_centre_y) =
                       fixed_values->Element(diamond_centre_x, diamond_centre_y).value;

                     overridden_by_fixed_array = true;
                  }
               }

               if (overridden_by_fixed_array == false)
               {
                  try
                  {
                     // Grab the values from the corners of the diamond. (Above, below, left, right)
                     float recalled_value = output_array.Element(diamond_centre_x               , diamond_centre_y + square_hsize)
                                          + output_array.Element(diamond_centre_x               , diamond_centre_y - square_hsize)
                                          + output_array.Element(diamond_centre_x - square_hsize, diamond_centre_y               )
                                          + output_array.Element(diamond_centre_x + square_hsize, diamond_centre_y               );

                     recalled_value /= 4.0f;

                     // Distort this value.
                     float distortion = ((rnd() * random_range * 2.0f) - random_range);
                     float new_midpoint_value = recalled_value + distortion;

                     // Place it
                     output_array.Element(diamond_centre_x, diamond_centre_y) = new_midpoint_value;
                  }
                  catch (int c)
                  {
                     // Array access out of bounds. Ignore this one then.
                     continue;
                  }
               }
            }
         }
      }

      // Alter variables.
      square_size = square_size / 2;
      iteration++;

      // Breaking conditions
      if (square_size == 1) break;

      // Reduce deviation range
      random_range *= environment_parameters.deviation_multiplier;
   }
}

WrappingArray2D<float> BlahBlur(const WrappingArray2D<float> &input_array)
{
   WrappingArray2D<float> na(input_array.Width(), input_array.Height());

   for (unsigned int x = 0; x < input_array.Width(); x++)
   {
      for (unsigned int y = 0; y < input_array.Height(); y++)
      {
         na.Element(x, y) = (input_array.Element(x - 1, y + 0)
                           + input_array.Element(x + 1, y + 0)
                           + input_array.Element(x + 0, y - 1)
                           + input_array.Element(x + 0, y + 1)
                           + input_array.Element(x + 0, y + 0)) / 5.0f;
      }
   }

   return na;
}

struct SIGMOID_CONTEXT
{
   float subtract_from_input_first, then_scale_input;

   float a, b, c;

   SIGMOID_CONTEXT(
      float _subtract_from_input_first,
      float _then_scale_input,
      float _a,
      float _b,
      float _c)
         : subtract_from_input_first(_subtract_from_input_first),
           then_scale_input(_then_scale_input),
           a(_a),
           b(_b),
           c(_c)
   {
   }

};

template <typename T>
void SigmoidTransformElement(T &element, const void *_sigmoid_context)
{
   const SIGMOID_CONTEXT *sigmoid_context = (const SIGMOID_CONTEXT *)_sigmoid_context;

   element -= sigmoid_context->subtract_from_input_first;
   element *= sigmoid_context->then_scale_input;

   element = sigmoid_context->a / (1.0f + exp(sigmoid_context->b * (sigmoid_context->c - element)));
}

// These are the number of BOXES in the grid.
void RegenerateTerrainHeightmap(Array2D<float> &values,
                                const ENVIRONMENT_GENERATION_PARAMETERS &environment_generation_parameters,
                                const Matrix2DInterface<FIXED_VALUE_ARRAY_ENTRY> *fixed_values_array = NULL)
{
   values.SetAll(0.0f);
   GenerateDSFractalTerrain(values, environment_generation_parameters, fixed_values_array);

   if (environment_generation_parameters.radial_falloff_enable)
   {
      const unsigned int &NO_X_V = environment_generation_parameters.no_x_cells,
                         &NO_Y_V = environment_generation_parameters.no_y_cells;

      // Number of -PINS-
      // Splat o multiply!
      Array2D<float> multiply_range(NO_X_V+1, NO_Y_V+1);


      for (unsigned int x = 0; x < environment_generation_parameters.no_x_cells+1; x++)
      {
         for (unsigned int y = 0; y < environment_generation_parameters.no_y_cells+1; y++)
         {
#define MAX_DIAGONAL_D   sqrt(abs(0 - (NO_X_V/2)) * abs(0 - (NO_X_V/2))   \
                            + abs(0 - (NO_Y_V/2)) * abs(0 - (NO_Y_V/2)))

#define MAX_ORTHOGONAL_D (NO_X_V/2)

            float distance = sqrt(abs(x - (NO_X_V/2)) * abs(x - (NO_X_V/2))
                                + abs(y - (NO_Y_V/2)) * abs(y - (NO_Y_V/2)));

            float multiplier;

            if (distance > MAX_ORTHOGONAL_D)
            {
               multiplier = 0.0f;
            }
            else
            {
               multiplier = pow(1.0f - (distance / MAX_ORTHOGONAL_D), environment_generation_parameters.radial_falloff_value);
            }

            multiply_range.Element(x, y) = multiplier;
         }
      }

      values = Multiply(values, multiply_range);
   }

   if (environment_generation_parameters.sigmoid_curve_enable)
   {
      SIGMOID_CONTEXT sigmoid_context = SIGMOID_CONTEXT(0.0f,
                                                        1.0f,
                                                        environment_generation_parameters.sigmoid_curve_a_parameter_value,
                                                        environment_generation_parameters.sigmoid_curve_b_parameter_value,
                                                        environment_generation_parameters.sigmoid_curve_c_parameter_value);

      values.ForEachUnary(SigmoidTransformElement<float>, &sigmoid_context);
   }

   //values = ClampBoth(values, MIN_E, MAX_E);
}

std::vector<GLVertexType> cached_terrain_mesh;

void RenderCachedTerrain()
{
   glEnableClientState(GL_VERTEX_ARRAY);
   glEnableClientState(GL_COLOR_ARRAY);
   glEnableClientState(GL_NORMAL_ARRAY);
   glVertexPointer(3, GL_FLOAT, sizeof(GLVertexType), cached_terrain_mesh[0].position.values);
   glColorPointer( 4, GL_FLOAT, sizeof(GLVertexType), cached_terrain_mesh[0].  colour.values);
   glNormalPointer(   GL_FLOAT, sizeof(GLVertexType), cached_terrain_mesh[0].  normal.values);
   glDrawArrays(GL_TRIANGLES, 0, cached_terrain_mesh.size());
}

COLOUR ReturnTerrainColourPseudorealistic(float v)
{
   if (v > 0.95f) return COLOUR::UC( 60,220, 60);
   if (v > 0.90f) return COLOUR::UC( 60,195, 60);
   if (v > 0.80f) return COLOUR::UC( 80,150, 80);
   if (v > 0.60f) return COLOUR::UC(140,125,  0);
   if (v > 0.40f) return COLOUR::UC(120,105, 20);
   return COLOUR::UC(100, 85, 40);
}

COLOUR ReturnTerrainColourGreyscale(float v)
{
   return COLOUR(mmax(0.0f, mmin(1.0f, v)));
}

void CreateAndCacheTerrainMesh(Matrix2DInterface<float> &values,
                               const ENVIRONMENT_GENERATION_PARAMETERS &egp)
{
   cached_terrain_mesh.clear();

   const unsigned int &NO_X_V = egp.no_x_cells;
   const unsigned int &NO_Y_V = egp.no_y_cells;

   int DISP_X = NO_X_V + 1, DISP_Y = NO_Y_V + 1; /* Assuming square! */
   const float HSCALE = 1.0f / (((float)DISP_X) - 1.0f);

   for (int x_box = 0; x_box < NO_X_V; x_box++)
   {
      for (int y_box = 0; y_box < NO_Y_V; y_box++)
      {
         int x_vertex_l = x_box + 0;
         int x_vertex_r = x_box + 1;
         int y_vertex_f = y_box + 0;
         int y_vertex_n = y_box + 1;

         Vector3D vertex_ln = Vector3D(HSCALE * (x_vertex_l - NO_X_V / 2.0f), values.Element(x_vertex_l, y_vertex_n), HSCALE * (y_vertex_n - NO_Y_V / 2.0f));
         Vector3D vertex_rn = Vector3D(HSCALE * (x_vertex_r - NO_X_V / 2.0f), values.Element(x_vertex_r, y_vertex_n), HSCALE * (y_vertex_n - NO_Y_V / 2.0f));
         Vector3D vertex_lf = Vector3D(HSCALE * (x_vertex_l - NO_X_V / 2.0f), values.Element(x_vertex_l, y_vertex_f), HSCALE * (y_vertex_f - NO_Y_V / 2.0f));
         Vector3D vertex_rf = Vector3D(HSCALE * (x_vertex_r - NO_X_V / 2.0f), values.Element(x_vertex_r, y_vertex_f), HSCALE * (y_vertex_f - NO_Y_V / 2.0f));

         Vector3D vertex_normal_tri_1 = TriangleNormal(vertex_lf, vertex_rn, vertex_rf);
         Vector3D vertex_normal_tri_2 = TriangleNormal(vertex_lf, vertex_ln, vertex_rn);

         float ratio_lf = (values.Element(x_vertex_l, y_vertex_f) - egp.minimum_colour_height_value) / (egp.maximum_colour_height_value - egp.minimum_colour_height_value);
         float ratio_rf = (values.Element(x_vertex_r, y_vertex_f) - egp.minimum_colour_height_value) / (egp.maximum_colour_height_value - egp.minimum_colour_height_value);
         float ratio_ln = (values.Element(x_vertex_l, y_vertex_n) - egp.minimum_colour_height_value) / (egp.maximum_colour_height_value - egp.minimum_colour_height_value);
         float ratio_rn = (values.Element(x_vertex_r, y_vertex_n) - egp.minimum_colour_height_value) / (egp.maximum_colour_height_value - egp.minimum_colour_height_value);

         COLOUR c_lf, c_rf, c_ln, c_rn;

         if (egp.terrain_colour_number == 1)
         {
            c_lf = ReturnTerrainColourPseudorealistic(ratio_lf);
            c_rf = ReturnTerrainColourPseudorealistic(ratio_rf);
            c_ln = ReturnTerrainColourPseudorealistic(ratio_ln);
            c_rn = ReturnTerrainColourPseudorealistic(ratio_rn);
         }
         else
         if (egp.terrain_colour_number == 0)
         {
            c_lf = ReturnTerrainColourGreyscale(ratio_lf);
            c_rf = ReturnTerrainColourGreyscale(ratio_rf);
            c_ln = ReturnTerrainColourGreyscale(ratio_ln);
            c_rn = ReturnTerrainColourGreyscale(ratio_rn);
         }
         else
         if (egp.terrain_colour_number == 2)
         {
            c_lf =
            c_rf =
            c_ln =
            c_rn = COLOUR_WHITE;
         }

         GLfloatVector3D(vertex_normal_tri_1).Values();

         GLVertexType v_lf, v_rf, v_ln, v_rn;
         v_lf.position = GLfloatVector3D(vertex_lf); v_lf.colour = c_lf;
         v_rf.position = GLfloatVector3D(vertex_rf); v_rf.colour = c_rf;
         v_ln.position = GLfloatVector3D(vertex_ln); v_ln.colour = c_ln;
         v_rn.position = GLfloatVector3D(vertex_rn); v_rn.colour = c_rn;

         if ((vertex_lf.values[Y] < egp.water_level)
          && (vertex_rf.values[Y] < egp.water_level)
          && (vertex_ln.values[Y] < egp.water_level)
          && (vertex_rn.values[Y] < egp.water_level))
         {
            continue;
         }

         v_lf.normal = v_rf.normal = v_ln.normal = v_rn.normal = GLfloatVector3D(vertex_normal_tri_1);

         cached_terrain_mesh.push_back(v_lf);
         cached_terrain_mesh.push_back(v_rf);
         cached_terrain_mesh.push_back(v_rn);

         v_lf.normal = v_rf.normal = v_ln.normal = v_rn.normal = GLfloatVector3D(vertex_normal_tri_2);

         cached_terrain_mesh.push_back(v_lf);
         cached_terrain_mesh.push_back(v_rn);
         cached_terrain_mesh.push_back(v_ln);
      }
   }
}

void takescreenshot()
{
   const char *output_filename = NULL;

   char filename[1024] = {0};

   if (output_filename == NULL)
   {
      time_t t = time(NULL);
      strftime(filename, 1023, "screens\\%Y-%m-%d %H-%M-%S screenshot.tga", localtime(&t));
      output_filename = filename;
   }

   FILE *output_file = fopen(output_filename, "wb");

   const unsigned short width = DISPLAY_X_SIZE, height = DISPLAY_Y_SIZE;
   const unsigned char * width_char_ptr = (const unsigned char *)& width,
                       *height_char_ptr = (const unsigned char *)&height;

   unsigned char tga_header[18] =
     {0x00,                                     // No extra data
      0x00,                                     // No palette
      0x02,                                     // RGB image
      0x00, 0x00, 0x00, 0x00,                   // No palette data
      0x20,                                     // 32 bit colour...?
      0x00, 0x00, 0x00, 0x00,                   // Start at 0,0
       width_char_ptr[0],  width_char_ptr[1],   // Width
      height_char_ptr[0], height_char_ptr[1],   // Height
      24,                                       // 24 bits per pixel
      0x00};                                     // No bit flips};

   fwrite(tga_header, 18, 1, output_file);

   unsigned char r, g, b;

   GLvoid *imageData = malloc(width * height * 4);

   glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, imageData);

   for (unsigned int y = 0; y < height; y++)
   {
      for (unsigned int x = 0; x < width; x++)
      {
         r = (((unsigned char *)imageData)[2 + ((x + (y * width)) * 4)]);
         g = (((unsigned char *)imageData)[1 + ((x + (y * width)) * 4)]);
         b = (((unsigned char *)imageData)[0 + ((x + (y * width)) * 4)]);

         fprintf(output_file, "%c%c%c", r, g, b);
         fflush(output_file);
      }
   }

   fclose(output_file);
}

// -----------------------------------------------------------------------------

int main(int argc, char *argv[])
{
   // Start SDL. (video)
   if (SDL_Init(SDL_INIT_VIDEO) < 0)
   {
      // Output error.
      printf("Unable to init SDL: %s\n", SDL_GetError());
      return 1;
   }

   // Make sure SDL cleans up before exit
   atexit(SDL_Quit);

   // Create an OpenGL window.
   SDL_WM_SetCaption("Mathew Carr - SDL OpenGL", NULL);

   screensurface       = SDL_SetVideoMode(DISPLAY_X_SIZE,
                                          DISPLAY_Y_SIZE,
                                          COLOUR_DEPTH,
                                          SDL_OPENGL | SDL_HWSURFACE);

   if (!screensurface)
   {
      // Output error.
      printf("Unable to create video output window: %s\n", SDL_GetError());
      return 1;
   }

   // Enable GLEW!
   glewInit();

   // Set up OpenGL.
   // OpenGL can draw to the entire screen.
   glViewport(0, 0, DISPLAY_X_SIZE, DISPLAY_Y_SIZE);

   // Set the clearing colour
   glClearColor(0.0f, 0.4f, 0.7f, 1.0f);

   // Set the depth buffer clearing depth.
   glClearDepth(1.0f);

   // Only draw polygons that are closer than / equal distance to those already drawn.
   glDepthFunc(GL_LEQUAL);

   // Turn on the depth test.
   glEnable(GL_DEPTH_TEST);

   // Lights on
   glEnable(GL_NORMALIZE);
   glEnable(GL_COLOR_MATERIAL);
   glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

   bool toexit = false;

   ENVIRONMENT_GENERATION_PARAMETERS environment_generation_parameters;
   PRESENTATION_PARAMETERS presentation_parameters;

   environment_generation_parameters.n_value = 7;

   unsigned int new_no_cells = power_of_two(environment_generation_parameters.n_value);

   environment_generation_parameters.no_x_cells = new_no_cells;
   environment_generation_parameters.no_y_cells = new_no_cells;

   environment_generation_parameters.water_level = 0.0f;

   environment_generation_parameters.starting_deviation_range_value = 1.0f;
   environment_generation_parameters.deviation_multiplier = 0.5f;

   environment_generation_parameters.radial_falloff_enable = false;
   environment_generation_parameters.radial_falloff_value = 0.5f;

   environment_generation_parameters.sigmoid_curve_enable = false;
   environment_generation_parameters.sigmoid_curve_a_parameter_value = 1.0f;
   environment_generation_parameters.sigmoid_curve_b_parameter_value = 1.0f;
   environment_generation_parameters.sigmoid_curve_c_parameter_value = 0.3f;

   environment_generation_parameters.minimum_colour_height_value = 0.0f;
   environment_generation_parameters.maximum_colour_height_value = 1.0f;

   environment_generation_parameters.terrain_colour_number = 0;

   presentation_parameters.lighting_enabled = false;
   presentation_parameters.semitransparent_water_active = false;
   presentation_parameters.smooth_shaded_terrain_active = false;

   //WrappingArray2D<float> values(NO_X_V, NO_Y_V);
   Array2D<float> height_map_values(environment_generation_parameters.no_x_cells+1, environment_generation_parameters.no_y_cells+1);

   Array2D<FIXED_VALUE_ARRAY_ENTRY> fixed_height_map_values_array(environment_generation_parameters.no_x_cells+1, environment_generation_parameters.no_y_cells+1);

   //values(NO_X_V/2, NO_Y_V/2) = 10.0f;

   //RegenerateTerrainHeightmap(height_map_values, environment_generation_parameters, &fixed_height_map_values_array);
   //CreateAndCacheTerrainMesh(height_map_values, environment_generation_parameters);

   Camera3D default_camera, default_camera2, default_camera3, current_camera, future_camera;

   default_camera.position = Vector3D( 0.0f,  1.5f,  +1.0f);
   default_camera.lateral  = Vector3D( 1.0f,  0.0f,   0.0f);
   default_camera.up       = Vector3D( 0.0f,  1.0f,   0.0f);
   //default_camera.ahead    = Vector3D( 0.0f,  0.0f,  -1.0f);
   default_camera.ahead = Vector3D(0.0f, 0.0f, 0.0f) - default_camera.position;

   default_camera2.position = Vector3D( 0.0f,  1.5f,  +1.0f);
   default_camera2.lateral  = Vector3D( 1.0f,  0.0f,   0.0f);
   default_camera2.up       = Vector3D( 0.0f,  1.0f,   0.0f);
   //default_camera.ahead    = Vector3D( 0.0f,  0.0f,  -1.0f);
   default_camera2.ahead = Vector3D(0.0f, 1.0f, 0.0f) - default_camera2.position;

   default_camera3.position = Vector3D( 0.0f,  2.5f,   0.0f);
   default_camera3.lateral  = Vector3D( 1.0f,  0.0f,   0.0f);
   default_camera3.up       = Vector3D( 0.0f,  0.0f,  -1.0f);
   //default_camera.ahead    = Vector3D( 0.0f,  0.0f,  -1.0f);
   default_camera3.ahead = Vector3D(0.0f, 0.0f, 0.0f) - default_camera3.position;

   current_camera = Camera3D(default_camera);
   future_camera  = Camera3D(current_camera);

   Uint8 p_mouse_states, c_mouse_states;
   int p_mouse_x, p_mouse_y, c_mouse_x, c_mouse_y;

   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR );

   //glEnable(GL_CULL_FACE);
   //glCullFace(GL_FRONT);

   WizardBraceletFont::Initialise();

   bool luminous_water = false;
   bool gui_display = true;

   enum SIMULATION_MODE
   {
      SIMULATION_MODE_VIEW_ENVIRONMENT,
      SIMULATION_MODE_ENTER_SEED_STRING,
   };

   SIMULATION_MODE simulation_mode = SIMULATION_MODE_ENTER_SEED_STRING;

   unsigned int enter_seed_string_cursor_position = 0;
   char enter_seed_string_seed_string[NO_SEED_STRING_CHARACTERS];

   bool update_terrain = true;

   GLfloatVector3D lightposition;

   float light_angle = 0.0f;

   // Main loop.
   while (!toexit) {

//------------------------------------------------------------------------------
      // Start the stopwatch!
      MSDL_SetStartTime();

      // Work all the events.
      MSDL_UpdateEventsArray();
      // Update the MSDL array of keystates.
      MSDL_GetKeyState();


      switch (simulation_mode)
      {
         case (SIMULATION_MODE_VIEW_ENVIRONMENT) :
         {
            if (MSDL_keystates[SDLK_h])
            {
               environment_generation_parameters.deviation_multiplier /= 1.01f;
               update_terrain = true;
            }
            if (MSDL_keystates[SDLK_j])
            {
               environment_generation_parameters.deviation_multiplier *= 1.01f;
               update_terrain = true;
            }

            if (MSDL_keystates[SDLK_n])
            {
               environment_generation_parameters.radial_falloff_value /= 1.02f;
               update_terrain = true;
            }
            if (MSDL_keystates[SDLK_m])
            {
               environment_generation_parameters.radial_falloff_value *= 1.02f;
               update_terrain = true;
            }
            if (MSDL_keystates[SDLK_v])
            {
               FIXED_VALUE_ARRAY_ENTRY &peak_element = fixed_height_map_values_array.Element(
                                                            environment_generation_parameters.no_x_cells / 2,
                                                            environment_generation_parameters.no_y_cells / 2);

               peak_element.value = (peak_element.value / 1.02f) - 0.005f;

               if (peak_element.value < 0.0f) peak_element.value = 0.0f;
               update_terrain = true;
            }
            if (MSDL_keystates[SDLK_b])
            {
               FIXED_VALUE_ARRAY_ENTRY &peak_element = fixed_height_map_values_array.Element(
                                                            environment_generation_parameters.no_x_cells / 2,
                                                            environment_generation_parameters.no_y_cells / 2);

               peak_element.value = (peak_element.value * 1.02f) + 0.005f;
               update_terrain = true;
            }

            if (MSDL_keystates[SDLK_i])
            {
               environment_generation_parameters.sigmoid_curve_b_parameter_value -= 0.21f;

               if (environment_generation_parameters.sigmoid_curve_b_parameter_value < 0.0f)
                   environment_generation_parameters.sigmoid_curve_b_parameter_value = 0.0f;
               update_terrain = true;
            }
            if (MSDL_keystates[SDLK_o])
            {
               environment_generation_parameters.sigmoid_curve_b_parameter_value += 0.21f;
               update_terrain = true;
            }

            if (MSDL_keystates[SDLK_k])
            {
               environment_generation_parameters.sigmoid_curve_c_parameter_value -= 0.04f;

               update_terrain = true;
            }
            if (MSDL_keystates[SDLK_l])
            {
               environment_generation_parameters.sigmoid_curve_c_parameter_value += 0.04f;

               update_terrain = true;
            }

            if (MSDL_keystates[SDLK_z])
            {
               environment_generation_parameters.water_level -= 0.005f;

               if (environment_generation_parameters.water_level < 0.0f)
                   environment_generation_parameters.water_level = 0.0f;

               environment_generation_parameters.minimum_colour_height_value = environment_generation_parameters.water_level;

               update_terrain = true;
            }
            if (MSDL_keystates[SDLK_x])
            {
               environment_generation_parameters.water_level += 0.005f;

               environment_generation_parameters.minimum_colour_height_value = environment_generation_parameters.water_level;

               update_terrain = true;
            }
            if (MSDL_KeyPressed(SDLK_TAB))
            {
               if (++environment_generation_parameters.seed_string[0] == 0) environment_generation_parameters.seed_string[1]++;
               update_terrain = true;
            }
            if (MSDL_KeyPressed(SDLK_DELETE))
            {
               gui_display = !gui_display;
            }
            if (MSDL_KeyPressed(SDLK_1))
            {
               environment_generation_parameters.radial_falloff_enable = !environment_generation_parameters.radial_falloff_enable;
               update_terrain = true;
            }
            if (MSDL_KeyPressed(SDLK_2))
            {
               environment_generation_parameters.sigmoid_curve_enable = !environment_generation_parameters.sigmoid_curve_enable;
               update_terrain = true;
            }
            if (MSDL_KeyPressed(SDLK_3))
            {
               FIXED_VALUE_ARRAY_ENTRY &peak_element = fixed_height_map_values_array.Element(
                                                            environment_generation_parameters.no_x_cells / 2,
                                                            environment_generation_parameters.no_y_cells / 2);

               peak_element.fixed = !peak_element.fixed;
               update_terrain = true;
            }
            if (MSDL_KeyPressed(SDLK_4))
            {
               presentation_parameters.semitransparent_water_active = !presentation_parameters.semitransparent_water_active;
            }
            if (MSDL_KeyPressed(SDLK_5))
            {
               presentation_parameters.smooth_shaded_terrain_active = !presentation_parameters.smooth_shaded_terrain_active;
            }
            if (MSDL_KeyPressed(SDLK_6))
            {
               presentation_parameters.lighting_enabled = !presentation_parameters.lighting_enabled;
            }
            if (MSDL_KeyPressed(SDLK_F1))
            {
               future_camera = Camera3D(default_camera);
            }
            if (MSDL_KeyPressed(SDLK_F2))
            {
               future_camera = Camera3D(default_camera2);
            }
            if (MSDL_KeyPressed(SDLK_F3))
            {
               future_camera = Camera3D(default_camera3);
            }
            if (MSDL_keystates[SDLK_F9])
            {
               future_camera.up = Vector3D(0.0f, 1.0f, 0.0f);
            }
            if (MSDL_keystates[SDLK_F11])
            {
               future_camera.ahead = Vector3D(0.0f, 1.0f, 0.0f) - future_camera.position;
               future_camera.RegainSanityAfterChange();
            }
            if (MSDL_keystates[SDLK_F12])
            {
               future_camera.ahead = Vector3D(0.0f, 0.0f, 0.0f) - future_camera.position;
               future_camera.RegainSanityAfterChange();
            }
            if (MSDL_KeyPressed(SDLK_END))
            {
               luminous_water = !luminous_water;
            }
            if (MSDL_KeyPressed(SDLK_PAGEDOWN))
            {
               if (++environment_generation_parameters.terrain_colour_number == 3) environment_generation_parameters.terrain_colour_number = 0;
               update_terrain = true;
            }
            if (MSDL_KeyPressed(SDLK_MINUS))
            {
               environment_generation_parameters.n_value -= 1;
               if (environment_generation_parameters.n_value < 1) environment_generation_parameters.n_value = 1;
               unsigned int new_no_cells = power_of_two(environment_generation_parameters.n_value);

               float peak_kicker_storage_value = fixed_height_map_values_array.Element(environment_generation_parameters.no_x_cells/2,
                                                                                       environment_generation_parameters.no_y_cells/2).value;

               bool  peak_kicker_storage_bool  = fixed_height_map_values_array.Element(environment_generation_parameters.no_x_cells/2,
                                                                                       environment_generation_parameters.no_y_cells/2).fixed;

               environment_generation_parameters.no_x_cells = new_no_cells;
               environment_generation_parameters.no_y_cells = new_no_cells;

               height_map_values = Array2D<float>(new_no_cells+1, new_no_cells+1);
               fixed_height_map_values_array = Array2D<FIXED_VALUE_ARRAY_ENTRY>(new_no_cells+1, new_no_cells+1);

               fixed_height_map_values_array.Element(environment_generation_parameters.no_x_cells/2,
                                                     environment_generation_parameters.no_y_cells/2).value = peak_kicker_storage_value;

               fixed_height_map_values_array.Element(environment_generation_parameters.no_x_cells/2,
                                                     environment_generation_parameters.no_y_cells/2).fixed = peak_kicker_storage_bool;

               update_terrain = true;
            }
            if (MSDL_KeyPressed(SDLK_EQUALS))
            {
               environment_generation_parameters.n_value += 1;
               unsigned int new_no_cells = power_of_two(environment_generation_parameters.n_value);

               float peak_kicker_storage_value = fixed_height_map_values_array.Element(environment_generation_parameters.no_x_cells/2,
                                                                                       environment_generation_parameters.no_y_cells/2).value;

               bool  peak_kicker_storage_bool  = fixed_height_map_values_array.Element(environment_generation_parameters.no_x_cells/2,
                                                                                       environment_generation_parameters.no_y_cells/2).fixed;

               environment_generation_parameters.no_x_cells = new_no_cells;
               environment_generation_parameters.no_y_cells = new_no_cells;

               height_map_values = Array2D<float>(new_no_cells+1, new_no_cells+1);
               fixed_height_map_values_array = Array2D<FIXED_VALUE_ARRAY_ENTRY>(new_no_cells+1, new_no_cells+1);

               fixed_height_map_values_array.Element(environment_generation_parameters.no_x_cells/2,
                                                     environment_generation_parameters.no_y_cells/2).value = peak_kicker_storage_value;

               fixed_height_map_values_array.Element(environment_generation_parameters.no_x_cells/2,
                                                     environment_generation_parameters.no_y_cells/2).fixed = peak_kicker_storage_bool;


               update_terrain = true;
            }
            if (MSDL_KeyPressed(SDLK_r))
            {
               simulation_mode = SIMULATION_MODE_ENTER_SEED_STRING;
               update_terrain = true;
               break;
            }

            if (update_terrain)
            {
               RegenerateTerrainHeightmap(height_map_values, environment_generation_parameters, &fixed_height_map_values_array);
               CreateAndCacheTerrainMesh(height_map_values, environment_generation_parameters);

               update_terrain = false;
            }

            if (MSDL_triggered_event_flags[SDL_QUIT] || (MSDL_keystates[SDLK_ESCAPE] == 1)) {
               toexit = true;
            }

            if (MSDL_KeyPressed(SDLK_g))
            {
               if (MSDL_keystates[SDLK_LSHIFT] || MSDL_keystates[SDLK_RSHIFT])
               {
                  RESCALE_REAL_CONTEXT rescale_real_context(environment_generation_parameters.water_level, 1.0f, 0.0f, 1.0f);

                  ExportToTGARescale(height_map_values, rescale_real_context, NULL);
               }
               else
               {
                  ExportToTGA(height_map_values, NULL);
               }
            }

            MSDL_GetMouseState(&c_mouse_states, &c_mouse_x, &c_mouse_y);

            int x_mouse_motion = c_mouse_x - p_mouse_x;
            int y_mouse_motion = c_mouse_y - p_mouse_y;

            const float MOUSE_MOTION_CAMERA_YAW_TURN_MULTIPLIER   = 0.0040f;
            const float MOUSE_MOTION_CAMERA_PITCH_TURN_MULTIPLIER = 0.0040f;
            const float MOUSE_MOTION_CAMERA_ROLL_TURN_MULTIPLIER  = 0.0040f;
            const float MOUSE_MOTION_CAMERA_ADVANCE_GO_MULTIPLIER = 0.0007f;

            const float KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME = 0.025f;

            if (MSDL_keystates[SDLK_a])
            {
               future_camera.position -= future_camera.lateral * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
            }
            if (MSDL_keystates[SDLK_d])
            {
               future_camera.position += future_camera.lateral * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
            }
            if (MSDL_keystates[SDLK_w])
            {
               future_camera.position += future_camera.ahead * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
            }
            if (MSDL_keystates[SDLK_s])
            {
               future_camera.position -= future_camera.ahead * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
            }
            if (MSDL_keystates[SDLK_q])
            {
               future_camera.position += future_camera.up * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
            }
            if (MSDL_keystates[SDLK_e])
            {
               future_camera.position -= future_camera.up * KEY_MOTION_CAMERA_MOVE_SPEED_PER_FRAME;
            }

            if (future_camera.position.values[Y] < 0.0f) future_camera.position.values[Y] = 0.0f;


            if (SDL_BUTTON(1) & c_mouse_states)
            {
               if (SDL_BUTTON(3) & c_mouse_states)
               {
                  future_camera.up += current_camera.lateral * (((float)x_mouse_motion) * MOUSE_MOTION_CAMERA_ROLL_TURN_MULTIPLIER);
               }
               else
               {
                  future_camera.ahead += current_camera.up * (((float)-y_mouse_motion) * MOUSE_MOTION_CAMERA_PITCH_TURN_MULTIPLIER);
                  future_camera.ahead += current_camera.lateral * (((float)x_mouse_motion) * MOUSE_MOTION_CAMERA_YAW_TURN_MULTIPLIER);
               }
            }
            else
            {
               if (SDL_BUTTON(3) & c_mouse_states)
               {
                  future_camera.position += current_camera.ahead * (((float)-y_mouse_motion) * MOUSE_MOTION_CAMERA_ADVANCE_GO_MULTIPLIER);
               }
            }

            future_camera.RegainSanityAfterChange();

            current_camera = current_camera.TendingToward(&future_camera, 0.2f);

            p_mouse_states = c_mouse_states;
            p_mouse_x = c_mouse_x;
            p_mouse_y = c_mouse_y;

            //------------------------------------------------------------------------

               // Camera.

               // Edit the projection matrix - set it to identity - set it to a perspective matrix.
               glMatrixMode(GL_PROJECTION);
               glLoadIdentity();
               gluPerspective(37.0f, ASPECT_RATIO, 0.01f, 30.01f);

               // Edit the modelview matrix - set it to identity.
               glMatrixMode(GL_MODELVIEW);
               glLoadIdentity();

               Matrix4D camera_matrix = current_camera.AsMatrix();


               glLoadMatrixf(camera_matrix.cell);

               if (presentation_parameters.lighting_enabled)
               {
                  light_angle += 0.01f;

                  glEnable(GL_LIGHTING);
                  glEnable(GL_LIGHT0);

                  Matrix4D yr; yr.SetToRotationY(light_angle);
                  lightposition = GLfloatVector3D(((Vector3D(0.0f, 1.0f * sin(light_angle * 0.4f) + 2.0f, 2.0f) * yr)));

                  glLightfv(GL_LIGHT0, GL_POSITION, lightposition.Values());

                  const float AMBIENT_LIGHT = 0.0f;
                  GLfloat ambient[4] = {AMBIENT_LIGHT, AMBIENT_LIGHT, AMBIENT_LIGHT, 1.0f};
                  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

                  GLfloat la[4] = {0.6f, 0.5f, 0.4f, 1.0f};
                  GLfloat ld[4] = {1.0f, 1.0f, 1.0f, 1.0f};
                  GLfloat lc[4] = {1.0f, 1.0f, 1.0f, 1.0f};
                  GLfloat ls[4] = {1.0f, 1.0f, 1.0f, 1.0f};

                  glLightfv(GL_LIGHT0, GL_AMBIENT,  la);
                  glLightfv(GL_LIGHT0, GL_DIFFUSE,  ld);
                  glLightfv(GL_LIGHT0, GL_COLOR,    lc);
               }
               else
               {
                  glDisable(GL_LIGHTING);
               }

               // Clear screen and depth buffer.
               glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

               glEnable(GL_DEPTH_TEST);
               glDepthMask(GL_TRUE);
               glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
               glBindTexture(GL_TEXTURE_2D, 0);
               glDisable(GL_TEXTURE_2D);
               glDisable(GL_BLEND);

               const float WATER_PLANE_HSIZE = 2.0f;
                  glBegin(GL_TRIANGLES);

                     GLVertexType v_lf = {Vector3D(-WATER_PLANE_HSIZE, environment_generation_parameters.water_level, -WATER_PLANE_HSIZE), COLOUR::UC(luminous_water?255:50, 107, 140)};
                     GLVertexType v_rf = {Vector3D( WATER_PLANE_HSIZE, environment_generation_parameters.water_level, -WATER_PLANE_HSIZE), COLOUR::UC(luminous_water?255:50, 157, 225)};
                     GLVertexType v_ln = {Vector3D(-WATER_PLANE_HSIZE, environment_generation_parameters.water_level,  WATER_PLANE_HSIZE), COLOUR::UC(luminous_water?255:50, 107, 140)};
                     GLVertexType v_rn = {Vector3D( WATER_PLANE_HSIZE, environment_generation_parameters.water_level,  WATER_PLANE_HSIZE), COLOUR::UC(luminous_water?255:50, 157, 225)};

                     glNormal3f(0.0f, 1.0f, 0.0f);

                     glColor4fv(v_lf.colour.values);
                     glVertex3fv(v_lf.position.values);
                     glColor4fv(v_rf.colour.values);
                     glVertex3fv(v_rf.position.values);
                     glColor4fv(v_rn.colour.values);
                     glVertex3fv(v_rn.position.values);

                     glColor4fv(v_lf.colour.values);
                     glVertex3fv(v_lf.position.values);
                     glColor4fv(v_ln.colour.values);
                     glVertex3fv(v_ln.position.values);
                     glColor4fv(v_rn.colour.values);
                     glVertex3fv(v_rn.position.values);
                  glEnd();

               if (presentation_parameters.smooth_shaded_terrain_active)
               {
                  // Smooth shading model
                  glShadeModel(GL_SMOOTH);
               }
               else
               {
                  // FLAT shading model
                  glShadeModel(GL_FLAT);
               }

               RenderCachedTerrain();

            glMatrixMode(GL_PROJECTION);
            glLoadIdentity();
            gluOrtho2D(ORTHO_PROJECTION_SCREEN_L, ORTHO_PROJECTION_SCREEN_R, ORTHO_PROJECTION_SCREEN_B, ORTHO_PROJECTION_SCREEN_T);
            glMatrixMode(GL_MODELVIEW);
            glLoadIdentity();

            glDisable(GL_DEPTH_TEST);

            glEnable(GL_BLEND);
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

            if (gui_display)
            {
               glDisable(GL_LIGHTING);

               char infoburst[20][200];

               sprintf(infoburst[0], "Camera Position: %6.2f, %6.2f, %6.2f", current_camera.position.values[X],
                                                                             current_camera.position.values[Y],
                                                                             current_camera.position.values[Z]);
               sprintf(infoburst[1], "   Ahead Vector: %6.2f, %6.2f, %6.2f", current_camera.ahead.values[X],
                                                                             current_camera.ahead.values[Y],
                                                                             current_camera.ahead.values[Z]);
               sprintf(infoburst[2], " Lateral Vector: %6.2f, %6.2f, %6.2f", current_camera.lateral.values[X],
                                                                             current_camera.lateral.values[Y],
                                                                             current_camera.lateral.values[Z]);
               sprintf(infoburst[3], " Upwards Vector: %6.2f, %6.2f, %6.2f", current_camera.up.values[X],
                                                                             current_camera.up.values[Y],
                                                                             current_camera.up.values[Z]);
               sprintf(infoburst[4], "Cached opengl vertices: %d", cached_terrain_mesh.size());
               sprintf(infoburst[19],"   Light Vector: %6.2f, %6.2f, %6.2f", lightposition.values[X],
                                                                             lightposition.values[Y],
                                                                             lightposition.values[Z]);

               sprintf(infoburst[5], "Z-X: Water level: %f", environment_generation_parameters.water_level);
               sprintf(infoburst[6], "K-L: Sigmoid curve C parameter: %f %s -2-", environment_generation_parameters.sigmoid_curve_c_parameter_value, environment_generation_parameters.sigmoid_curve_enable ? "on" : "off");
               sprintf(infoburst[7], "I-O: Sigmoid curve B parameter: %f %s -2-", environment_generation_parameters.sigmoid_curve_b_parameter_value, environment_generation_parameters.sigmoid_curve_enable ? "on" : "off");
               sprintf(infoburst[8], "N-M: Radial Falloff: %f %s -1-", environment_generation_parameters.radial_falloff_value, environment_generation_parameters.radial_falloff_enable ? "on" : "off");
               sprintf(infoburst[9], "H-J: Algorithm jagged falloff: %f", environment_generation_parameters.deviation_multiplier);
               sprintf(infoburst[10],"V-B: Peak kicker: %f %s -3-", fixed_height_map_values_array.Element(environment_generation_parameters.no_x_cells/2,environment_generation_parameters.no_y_cells/2).value,
                                                                    fixed_height_map_values_array.Element(environment_generation_parameters.no_x_cells/2,environment_generation_parameters.no_y_cells/2).fixed ? "on" : "off");

               WizardBraceletFont::DrawString(infoburst[0],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_T - 0 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(255, 255, 255));
               WizardBraceletFont::DrawString(infoburst[1],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_T - 1 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(155, 255, 255));
               WizardBraceletFont::DrawString(infoburst[2],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_T - 2 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(255, 155, 255));
               WizardBraceletFont::DrawString(infoburst[3],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_T - 3 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(255, 255, 155));
               WizardBraceletFont::DrawString(infoburst[4],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_T - 5 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(255, 255, 255));
               WizardBraceletFont::DrawString(infoburst[19],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_T - 7 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(255, 255, 255));
               WizardBraceletFont::DrawString(infoburst[5],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_B + 1 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(205, 225, 255));
               WizardBraceletFont::DrawString(infoburst[6],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_B + 2 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(255, environment_generation_parameters.sigmoid_curve_enable ? 255:100, environment_generation_parameters.sigmoid_curve_enable ? 255:100));
               WizardBraceletFont::DrawString(infoburst[7],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_B + 3 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(255, environment_generation_parameters.sigmoid_curve_enable ? 255:100, environment_generation_parameters.sigmoid_curve_enable ? 255:100));
               WizardBraceletFont::DrawString(infoburst[8],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_B + 4 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(255, environment_generation_parameters.radial_falloff_enable ? 255:100, environment_generation_parameters.radial_falloff_enable ? 255:100));
               WizardBraceletFont::DrawString(infoburst[9],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_B + 5 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(255, 225, 225));
               WizardBraceletFont::DrawString(infoburst[10],
                                              ORTHO_PROJECTION_SCREEN_L,
                                              ORTHO_PROJECTION_SCREEN_B + 6 * WizardBraceletFont::DEFAULT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              WizardBraceletFont::DEFAULT_CHARACTER_SIZE,
                                              COLOUR::UC(255, 225, 225));

               const float STRING_CHAR_OUTPUT_CHARACTER_SIZE = 0.15f;

               char seed_string_output[1000] = {0};

               char *seed_string_output_ptr = seed_string_output;

               const unsigned int SEED_STRING_CHAR_OUTPUT = 20;

               *seed_string_output_ptr++ = '\"';

               for (unsigned int seed_string_i = 0; seed_string_i < SEED_STRING_CHAR_OUTPUT; seed_string_i++)
               {
                  const unsigned char &seed_character = environment_generation_parameters.seed_string[seed_string_i];

                  if (seed_character > 32)
                  {
                     *seed_string_output_ptr++ = seed_character;
                  }
                  else
                  {
                     *seed_string_output_ptr++ = ' ';
                  }
               }

               *seed_string_output_ptr++ = '\"';

               *seed_string_output_ptr++ = '\n';
               *seed_string_output_ptr++ = '\n';
               *seed_string_output_ptr++ = ' ';

               for (unsigned int seed_string_i = 0; seed_string_i < NO_SEED_STRING_CHARACTERS; seed_string_i++)
               {
                  const unsigned char &seed_character = environment_generation_parameters.seed_string[seed_string_i];

                  char hex[3];

                  sprintf(hex, "%02X", seed_character);

                  *seed_string_output_ptr++ = hex[0];
                  *seed_string_output_ptr++ = hex[1];

                  if (((seed_string_i + 1) % 10) == 0)
                  {
                     *seed_string_output_ptr++ = '\n';
                     *seed_string_output_ptr++ = ' ';
                  }
               }

               *seed_string_output_ptr++ = 0;

               WizardBraceletFont::DrawString(seed_string_output,
                                              ORTHO_PROJECTION_SCREEN_R - 24 * STRING_CHAR_OUTPUT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_CHARACTER_SPACING,
                                              ORTHO_PROJECTION_SCREEN_B + 14 * STRING_CHAR_OUTPUT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                              STRING_CHAR_OUTPUT_CHARACTER_SIZE,
                                              STRING_CHAR_OUTPUT_CHARACTER_SIZE,
                                              COLOUR::UC(255, 255, 205));

            }

            glBindTexture(GL_TEXTURE_2D, 0);

            glDisable(GL_BLEND);
         }
         break;

         case (SIMULATION_MODE_ENTER_SEED_STRING) :
         {
            /* printable characters */
            if (enter_seed_string_cursor_position != NO_SEED_STRING_CHARACTERS)
            {
               for (unsigned int test_key = SDLK_SPACE; test_key < SDLK_DELETE; test_key++)
               {
                  if (MSDL_KeyPressed(test_key))
                  {
                     unsigned int update_key = test_key;

                     if ((MSDL_keystates[SDLK_LSHIFT] || MSDL_keystates[SDLK_RSHIFT]) && (test_key >= SDLK_a) && (test_key <= SDLK_z))
                     {
                        update_key += 'A' - 'a';
                     }

                     enter_seed_string_seed_string[enter_seed_string_cursor_position] = update_key;
                     enter_seed_string_cursor_position++;
                  }
               }
            }

            if (MSDL_KeyPressed(SDLK_BACKSPACE))
            {
               if (enter_seed_string_cursor_position > 0) enter_seed_string_cursor_position--;
            }

            if (MSDL_KeyPressed(SDLK_RETURN))
            {
               simulation_mode = SIMULATION_MODE_VIEW_ENVIRONMENT;

               for (unsigned int seed_string_character = 0; seed_string_character < NO_SEED_STRING_CHARACTERS; seed_string_character++)
               {
                  if (seed_string_character < enter_seed_string_cursor_position)
                  {
                     environment_generation_parameters.seed_string[seed_string_character] = enter_seed_string_seed_string[seed_string_character];
                  }
                  else
                  {
                     environment_generation_parameters.seed_string[seed_string_character] = 0;
                  }
               }
            }

            if (MSDL_triggered_event_flags[SDL_QUIT] || (MSDL_keystates[SDLK_ESCAPE] == 1)) {
               toexit = true;
            }
               glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

               // Edit the projection matrix - set it to identity - set it to a perspective matrix.
               glMatrixMode(GL_PROJECTION);
               glLoadIdentity();
               gluPerspective(37.0f, ASPECT_RATIO, 0.01f, 30.01f);

               // Edit the modelview matrix - set it to identity.
               glMatrixMode(GL_MODELVIEW);
               glLoadIdentity();


            glMatrixMode(GL_PROJECTION);
            glLoadIdentity();
            gluOrtho2D(ORTHO_PROJECTION_SCREEN_L, ORTHO_PROJECTION_SCREEN_R, ORTHO_PROJECTION_SCREEN_B, ORTHO_PROJECTION_SCREEN_T);
            glMatrixMode(GL_MODELVIEW);
            glLoadIdentity();

            glDisable(GL_DEPTH_TEST);

            glEnable(GL_BLEND);
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

            const float ENTER_SEED_STRING_TITLE_PROMPT_CHARACTER_SIZE = 0.4f;
            const float ENTER_SEED_STRING_SEED_STRING_TEXT_CHARACTER_SIZE = 0.6f;

            WizardBraceletFont::DrawStringCentered("Enter New Seed String:",
                                                   0.0f,
                                                   ORTHO_PROJECTION_SCREEN_T - 1 * ENTER_SEED_STRING_TITLE_PROMPT_CHARACTER_SIZE * WizardBraceletFont::DEFAULT_LINE_SPACING,
                                                   ENTER_SEED_STRING_TITLE_PROMPT_CHARACTER_SIZE,
                                                   ENTER_SEED_STRING_TITLE_PROMPT_CHARACTER_SIZE);

            char seed_string_cursor_display[NO_SEED_STRING_CHARACTERS * 2] = {0};
            char *seed_string_cursor_display_ptr = seed_string_cursor_display;

            for (unsigned int current_character = 0; current_character < enter_seed_string_cursor_position; current_character++)
            {
               const unsigned int SEED_STRING_CURSOR_DISPLAY_CHARACTERS_PER_LINE = 22;

               *seed_string_cursor_display_ptr++ = enter_seed_string_seed_string[current_character];

               if (((current_character + 1) % SEED_STRING_CURSOR_DISPLAY_CHARACTERS_PER_LINE) == 0)
               {
                  *seed_string_cursor_display_ptr++ = '\n';
               }
            }

            *seed_string_cursor_display_ptr++ = '_';

            WizardBraceletFont::DrawString(seed_string_cursor_display,
                                           ORTHO_PROJECTION_SCREEN_L,
                                           0.0f,
                                           ENTER_SEED_STRING_SEED_STRING_TEXT_CHARACTER_SIZE,
                                           ENTER_SEED_STRING_SEED_STRING_TEXT_CHARACTER_SIZE);

            glBindTexture(GL_TEXTURE_2D, 0);

            glDisable(GL_BLEND);
         }
         break;
      }

      // Swap backbuffer to the screen.
      SDL_GL_SwapBuffers();

      if (MSDL_KeyPressed(SDLK_HOME))
      {
         takescreenshot();
      }

      // Limit program speed to 60FPS.
      MSDL_SpinlockUntilStopwatch(15);
   }

//------------------------------------------------------------------------------

   WizardBraceletFont::Deinitialise();

   SDL_Quit();

   // End program.
   return 0;
}
